from __future__ import annotations

from typing import TYPE_CHECKING

from deepdiff import DeepDiff
from yarl import URL

from pylav.compat import json
from pylav.extension.bundled_node import LAVALINK_DOWNLOAD_DIR
from pylav.storage.migrations.logging import LOGGER

if TYPE_CHECKING:
    from pylav.core.client import Client


async def update_plugins(client: Client) -> None:
    """Update the plugins in the database."""
    try:
        LOGGER.info("Attempting to update plugins")
        # noinspection PyProtectedMember
        config = client._node_config_manager.bundled_node_config()
        data = await config.fetch_yaml()
        new_plugin_data = []
        _temp = set()
        existing_plugins = set()
        baseurl = URL("https://api.github.com/repos")
        for plugin in data["lavalink"]["plugins"].copy():
            dependency = ":".join(plugin["dependency"].split(":")[:-1])
            if dependency in _temp:
                continue
            _temp.add(dependency)
            if plugin["dependency"].startswith("com.github.TopiSenpai.LavaSrc:lavasrc-plugin:"):
                existing_plugins.add("lavasrc-plugin-")
                org = "TopiSenpai"
                repo = "LavaSrc"
                repository = "https://maven.topi.wtf/releases"
                dependency += ":"
            elif plugin["dependency"].startswith("com.dunctebot:skybot-lavalink-plugin:"):
                existing_plugins.add("skybot-lavalink-plugin-")
                org = "DuncteBot"
                repo = "skybot-lavalink-plugin"
                repository = "https://m2.duncte123.dev/releases"
                dependency += ":"
            elif plugin["dependency"].startswith("com.github.topisenpai:sponsorblock-plugin:"):
                existing_plugins.add("sponsorblock-plugin-")
                org = "Topis-Lavalink-Plugins"
                repo = "Sponsorblock-Plugin"
                repository = "https://jitpack.io"
                dependency += ":"
            elif plugin["dependency"].startswith("com.github.esmBot:lava-xm-plugin:"):
                existing_plugins.add("lava-xm-plugin-")
                org = "esmBot"
                repo = "lava-xm-plugin"
                repository = "https://jitpack.io"
                dependency += ":"
            elif plugin["dependency"].startswith("me.rohank05:lavalink-filter-plugin:"):
                existing_plugins.add("lavalink-filter-plugin-")
                org = "rohank05"
                repo = "lavalink-filter-plugin"
                repository = "https://jitpack.io"
                dependency += ":"
            else:
                continue
            release_data = await (await client.cached_session.get(baseurl / org / repo / "releases" / "latest")).json(
                loads=json.loads
            )
            name = release_data["tag_name"]
            new_plugin_data.append(
                {
                    "dependency": dependency + name,
                    "repository": repository,
                }
            )

        if __ := DeepDiff(
            data["lavalink"]["plugins"], new_plugin_data, ignore_order=True, max_passes=3, cache_size=10000
        ):
            data["lavalink"]["plugins"] = new_plugin_data
            LOGGER.info("New plugin version: %s", new_plugin_data)
            await config.update_yaml(data)
        else:
            LOGGER.info("No plugin updates required")

        folder = LAVALINK_DOWNLOAD_DIR / "plugins"
        if not await folder.exists():
            return
        async for file in folder.iterdir():
            if (
                await file.is_file()
                and file.name.endswith(".jar")
                and not file.name.startswith(tuple(existing_plugins))
            ):
                try:
                    await file.unlink()
                    LOGGER.warning("Removed old plugin: %s", file.name)
                except Exception as exc:
                    LOGGER.error("Failed to remove old plugin: %s", file.name, exc_info=exc)
    except Exception as exc:
        LOGGER.error("Failed to update plugins", exc_info=exc)



import contextlib
import json
from collections.abc import Callable
from json import JSONDecodeError as JSONDecodeError
from json import JSONDecoder as JSONDecoder
from json import JSONEncoder as JSONEncoder
from types import ModuleType
from typing import IO, Any, AnyStr, overload

try:
    import orjson as _orjson

except ImportError:
    _orjson = None

try:
    import ujson as _ujson
except ImportError:
    _ujson = None


__all__ = [
    "dump",
    "dumps",
    "load",
    "loads",
    "JSONDecoder",
    "JSONDecodeError",
    "JSONEncoder",
    "get_origin",
]

# .dumps() signatures
if _orjson:

    @overload
    def dumps(
        obj: Any,
        *,
        orjson_default: Callable[[Any], Any] | None = None,
        orjson_option: int | None = None,
    ) -> str:
        """Serialize ``obj`` to a JSON formatted ``str``."""
        ...

    __dumps_origin = _orjson

elif _ujson:

    @overload
    def dumps(
        obj: Any,
        *,
        ensure_ascii: bool = True,
        ujson_encode_html_chars: bool = False,
        ujson_escape_forward_slashes: bool = True,
        sort_keys: bool = False,
        indent: int | None = 0,
    ) -> str:
        """Serialize ``obj`` to a JSON formatted ``str``."""
        ...

    __dumps_origin = _ujson

else:

    @overload
    def dumps(
        obj: Any,
        *,
        skipkeys: bool = False,
        ensure_ascii: bool = True,
        check_circular: bool = True,
        allow_nan: bool = True,
        cls: type[JSONEncoder] | None = None,
        indent: None | int | str = None,
        separators: tuple[str, str] | None = None,
        default: Callable[[Any], Any] | None = None,
        sort_keys: bool = False,
        **kwds: Any,
    ) -> str:
        """Serialize ``obj`` to a JSON formatted ``str``."""
        ...

    __dumps_origin = json


# .loads() signatures
if _orjson:

    @overload
    def loads(obj: bytes | bytearray | memoryview | str) -> Any:
        """Deserialize ``obj`` (a ``str``, ``bytes`` or ``bytearray`` instance"""
        ...

    __loads_origin = _orjson

elif _ujson:

    @overload
    def loads(obj: AnyStr, *, ujson_precise_float: bool | None = None) -> Any:
        """Deserialize ``obj`` (a ``str``, ``bytes`` or ``bytearray`` instance"""
        ...

    __loads_origin = _ujson

else:

    @overload
    def loads(
        obj: AnyStr,
        *,
        cls: type[JSONDecoder] | None = None,
        object_hook: Callable[[dict[Any, Any]], Any] | None = None,
        parse_float: Callable[[str], Any] | None = None,
        parse_int: Callable[[str], Any] | None = None,
        parse_constant: Callable[[str], Any] | None = None,
        object_pairs_hook: Callable[[list[tuple[Any, Any]]], Any] | None = None,
        **kwargs: Any,
    ) -> Any:
        """Deserialize ``obj`` (a ``str``, ``bytes`` or ``bytearray`` instance"""
        ...

    __loads_origin = json


# .dump() signatures
if _orjson:

    @overload
    def dump(
        obj: Any,
        fp: IO[str],
    ) -> None:
        """Serialize ``obj`` as a JSON formatted stream to ``fp`` (a ``.write()``-supporting file-like object)."""
        ...

    __dump_origin = _orjson

elif _ujson:

    @overload
    def dump(
        obj: Any,
        fp: IO[str],
        *,
        ensure_ascii: bool = True,
        sort_keys: bool = False,
        indent: int = 0,
        ujson_encode_html_chars: bool = False,
        ujson_escape_forward_slashes: bool = True,
    ) -> None:
        """Serialize ``obj`` as a JSON formatted stream to ``fp`` (a ``.write()``-supporting file-like object)."""
        ...

    __dump_origin = _ujson

else:

    @overload
    def dump(
        obj: Any,
        fp: IO[str],
        *,
        skipkeys: bool = False,
        ensure_ascii: bool = True,
        check_circular: bool = True,
        allow_nan: bool = True,
        cls: type[JSONEncoder] | None = None,
        indent: None | int | str = None,
        separators: tuple[str, str] | None = None,
        default: Callable[[Any], Any] | None = None,
        sort_keys: bool = False,
        **kwargs: Any,
    ) -> None:
        """Serialize ``obj`` as a JSON formatted stream to ``fp`` (a ``.write()``-supporting file-like object)."""
        ...

    __dump_origin = json


# .load() signatures

if _orjson:

    @overload
    def load(fp: IO[AnyStr]) -> Any:
        """Deserialize ``fp`` (a ``.read()``-supporting file-like object containing a JSON document) to a Python object."""
        ...

    __load_origin = _orjson

elif _ujson:

    @overload
    def load(
        fp: IO[AnyStr],
        *,
        precise_float: bool = ...,
    ) -> Any:
        """Deserialize ``fp`` (a ``.read()``-supporting file-like object containing a JSON document) to a Python object."""
        ...

    __load_origin = _ujson

else:

    @overload
    def load(
        fp: IO[AnyStr],
        *,
        cls: type[JSONDecoder] | None = ...,
        object_hook: Callable[[Any], Any] | None = ...,
        parse_float: Callable[[str], float] | None = ...,
        parse_int: Callable[[str], int] | None = ...,
        parse_constant: Callable[[str], Any] | None = ...,
        object_pairs_hook: Callable[[list[tuple[str, Any]]], Any] | None = ...,
        **kwargs: Any,
    ) -> Any:
        """Deserialize ``fp`` (a ``.read()``-supporting file-like object containing a JSON document) to a Python object."""
        ...

    __load_origin = json


def dumps(
    obj: Any,
    *,
    skipkeys=False,
    ensure_ascii=True,
    check_circular=True,
    allow_nan=True,
    cls=None,
    indent=None,
    separators=None,
    default=None,
    sort_keys=False,
    orjson_default=None,
    orjson_option=None,
    ujson_encode_html_chars=False,
    ujson_escape_forward_slashes=True,
    **kwargs,
) -> str:
    """
    Serialize ``obj`` to a JSON formatted ``str``.

    Parameters
    ----------
    obj : Any
        The object to serialize.
    skipkeys : bool, optional
        If ``True`` (default: ``False``), keys that are not basic types (``str``, ``int``, ``float``, ``bool``, ``None``) will be skipped instead of raising a ``TypeError``.
    ensure_ascii: bool, optional
        If ``True`` (default: ``True``), the output is guaranteed to have all incoming non-ASCII characters escaped. If ``False``, these characters will be output as-is.
    check_circular: bool, optional
        If check_circular is ``False`` (default: ``True``), then the circular reference check for container types will be skipped and a circular reference will result in an ``RecursionError`` (or worse).
    allow_nan: bool, optional
        If ``False`` (default: ``True``), then it will be a ValueError to serialize out of range float values (`nan`, `inf`, `-inf`) in strict compliance of the JSON specification. If ``True``, their JavaScript equivalents (`NaN`, `Infinity`, `-Infinity`) will be used.
    cls : type, optional
        If specified, must be a subclass of ``json.JSONEncoder``. An instance is used to encode the object.
    indent : int or str, optional
        If specified, then JSON array elements and object members will be pretty-printed with a newline followed by that many spaces. An indent level of 0, negative, or ``None`` will only insert newlines. ``None`` is the most compact representation. Using a negative indent indents that many spaces after the ``newline`` character. If it is a string (such as ``'\t'``), that string is used to indent each level. Default: ``None``.
    separators : tuple, optional
        If specified, then it should be an (item_separator, key_separator) tuple. The default is ``(', ', ': ')`` if *indent* is ``None`` and ``(',', ': ')`` otherwise. To get the most compact JSON representation, you should specify ``(',', ':')`` to eliminate whitespace.
    default: Callable, optional
        If specified, then it should be a function that gets called for objects that can't otherwise be serialized. It should return a JSON encodable version of the object or raise a ``TypeError``. If not specified, ``TypeError`` is raised.
    sort_keys: bool, optional
        If ``True`` (default: ``False``), then the output of dictionaries will be sorted by key.
    orjson_default: Callable, optional
        If specified, then it should be a function that gets called for objects that can't otherwise be serialized. It should return a JSON encodable version of the object or raise a ``TypeError``. If not specified, ``TypeError`` is raised.
    orjson_option: int, optional
        If specified, then it should be an integer that is passed to ``orjson.dumps`` as the ``option`` parameter. If not specified, ``None`` is used.
    ujson_encode_html_chars: bool, optional
        If ``True`` (default: ``False``), then the output will have the characters ``<``, ``>``, ``&`` encoded as ``\u003c``, ``\u003e``, ``\u0026``.
    ujson_escape_forward_slashes:
        If ``True`` (default: ``True``), then the output will have the forward slash character ``/`` encoded as ``\\/``.
    kwargs: Any, optional
        Additional keyword arguments are passed to ``json.dumps``.

    Returns
    -------
    str
        The JSON string representation of ``obj``.

    """
    if _orjson:
        with contextlib.suppress(_orjson.JSONEncodeError):
            return _orjson.dumps(obj, default=orjson_default, option=orjson_option).decode()
    if _ujson:
        return _ujson.dumps(
            obj,
            ensure_ascii=ensure_ascii,
            sort_keys=sort_keys,
            indent=indent or 0,
            encode_html_chars=ujson_encode_html_chars,
            escape_forward_slashes=ujson_escape_forward_slashes,
        )
    return json.dumps(
        obj,
        skipkeys=skipkeys,
        ensure_ascii=ensure_ascii,
        check_circular=check_circular,
        allow_nan=allow_nan,
        cls=cls,
        indent=indent,
        separators=separators,
        default=default,
        sort_keys=sort_keys,
        **kwargs,
    )


def loads(
    obj: AnyStr | bytes | bytearray | memoryview | str,
    *,
    cls=None,
    object_hook=None,
    parse_float=None,
    parse_int=None,
    parse_constant=None,
    object_pairs_hook=None,
    ujson_precise_float=None,
    **kwargs,
) -> Any:
    """Deserialize ``obj`` (a ``str``, ``bytes`` or ``bytearray`` instance containing a JSON document) to a Python object.

    Parameters
    ----------
    obj: AnyStr | bytes | bytearray | memoryview | str
        The JSON string to deserialize.
    cls: type, optional
        If specified, must be a subclass of ``json.JSONDecoder``. An instance is used to decode the object.
    object_hook: Callable, optional
        If specified, then it should be a function that will be called with the result of any object literal decoded (a ``dict``). The return value of ``object_hook`` will be used instead of the ``dict``. This feature can be used to implement custom decoders (e.g. JSON-RPC class hinting).
    parse_float: Callable, optional
        If specified, then it should be a function to be called with the string of every JSON float to be decoded. By default this is equivalent to ``float(num_str)``. This can be used to use another datatype or parser for JSON floats (e.g. ``decimal.Decimal``).
    parse_int: Callable, optional
        If specified, then it should be a function to be called with the string of every JSON int to be decoded. By default this is equivalent to ``int(num_str)``. This can be used to use another datatype or parser for JSON integers (e.g. ``float``).
    parse_constant: Callable, optional
        If specified, then it should be a function to be called with one of the following strings: ``'-Infinity'``, ``'Infinity'``, ``'NaN'``. This can be used to raise an exception if invalid JSON numbers are encountered.
    object_pairs_hook: Callable, optional
        If specified, then it should be a function that will be called with the result of any object literal decoded with an ordered list of pairs. The return value of ``object_pairs_hook`` will be used instead of the ``dict``. This feature can be used to implement custom decoders. If ``object_hook`` is also defined, the ``object_pairs_hook`` takes priority.
    ujson_precise_float: bool, optional
        If ``True`` (default: ``False``), then
    kwargs: Any, optional
        Additional keyword arguments are passed to ``json.loads``.

    Returns
    -------
    Any
        The deserialized object.

    Raises
    ------
    json.JSONDecodeError
        If the input is not valid JSON.
    orjson.JSONDecodeError(json.JSONDecodeError, ValueError)
        If the input is not valid JSON and orjson is used.
    ValueError
        If the string is not correctly formed and ujson is used.

    """
    if _orjson:
        with contextlib.suppress(_orjson.JSONDecodeError):
            return _orjson.loads(obj)
    if _ujson:
        with contextlib.suppress(_ujson.JSONDecodeError):
            return _ujson.loads(obj, precise_float=ujson_precise_float)
    return json.loads(
        obj,
        cls=cls,
        object_hook=object_hook,
        parse_float=parse_float,
        parse_int=parse_int,
        parse_constant=parse_constant,
        object_pairs_hook=object_pairs_hook,
        **kwargs,
    )


def dump(
    obj: Any,
    fp: IO[AnyStr],
    *,
    skipkeys=False,
    ensure_ascii=True,
    check_circular=True,
    allow_nan=True,
    cls=None,
    indent=None,
    separators=None,
    default=None,
    sort_keys=False,
    ujson_encode_html_chars=False,
    ujson_escape_forward_slashes=True,
    **kwargs,
) -> None:
    """Serialize ``obj`` as a JSON formatted stream to ``fp`` (a ``.write()``-supporting file-like object)."""

    if _orjson:
        with contextlib.suppress(_orjson.JSONEncodeError):
            fp.write(_orjson.dumps(obj).decode())
            return
    if _ujson:
        return _ujson.dump(
            obj,
            fp,
            ensure_ascii=ensure_ascii,
            sort_keys=sort_keys,
            indent=indent or 0,
            encode_html_chars=ujson_encode_html_chars,
            escape_forward_slashes=ujson_escape_forward_slashes,
        )
    return json.dump(
        obj,
        fp,
        skipkeys=skipkeys,
        ensure_ascii=ensure_ascii,
        check_circular=check_circular,
        allow_nan=allow_nan,
        cls=cls,
        indent=indent,
        separators=separators,
        default=default,
        sort_keys=sort_keys,
        **kwargs,
    )


def load(
    fp: IO[AnyStr],
    *,
    cls=None,
    object_hook=None,
    parse_float=None,
    parse_int=None,
    parse_constant=None,
    object_pairs_hook=None,
    precise_float=None,
    **kwargs,
) -> Any:
    """Deserialize ``fp`` (a ``.read()``-supporting file-like object containing a JSON document) to a Python object."""
    if _orjson:
        with contextlib.suppress(_orjson.JSONDecodeError):
            return _orjson.loads(fp.read())
    if _ujson:
        with contextlib.suppress(_ujson.JSONDecodeError):
            return _ujson.load(fp, precise_float=precise_float)
    return json.load(
        fp,
        cls=cls,
        object_hook=object_hook,
        parse_float=parse_float,
        parse_int=parse_int,
        parse_constant=parse_constant,
        object_pairs_hook=object_pairs_hook,
        **kwargs,
    )


def get_origin() -> dict[str, ModuleType]:
    """Return a dict of json modules being used."""
    return {
        "dumps": __dumps_origin,
        "loads": __loads_origin,
        "dump": __dump_origin,
        "load": __load_origin,
    }



from __future__ import annotations

import contextlib
from pathlib import Path
from typing import Literal

import discord
from redbot.core.i18n import Translator

from pylav.helpers import emojis
from pylav.type_hints.bot import DISCORD_COG_TYPE, DISCORD_INTERACTION_TYPE

_ = Translator("PyLav", Path(__file__))


class DisconnectButton(discord.ui.Button):
    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        style: discord.ButtonStyle,
        row: int = None,
    ):
        super().__init__(
            style=style,
            emoji=emojis.POWER,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True)
        context = await self.cog.bot.get_context(interaction)
        if not await self.view.bot.is_owner(context.author):
            await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=context, title=_("You are not authorized to perform this action.")
                ),
                ephemeral=True,
            )
            return
        player = self.view.source.current_player
        if not player:
            await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=context, title=_("There are no players available for this action. Try refreshing.")
                ),
                ephemeral=True,
            )
            return
        if notify_channel := await player.notify_channel():
            with contextlib.suppress(discord.HTTPException):
                await notify_channel.send(
                    embed=await self.cog.pylav.construct_embed(
                        title=_("My bot owner requested an action"),
                        description=_(
                            "My bot owner has requested that I disconnect from this server. "
                            "I have saved my current position, so connect me to a voice channel to resume playback. "
                            "Sorry for any inconvenience caused."
                        ),
                    )
                )
        await player.disconnect(requester=context.author)

        await self.view.prepare()
        kwargs = await self.view.get_page(self.view.current_page)
        attachments = []
        if "file" in kwargs:
            attachments = [kwargs.pop("file")]
        elif "files" in kwargs:
            attachments = kwargs.pop("files")
        if attachments:
            kwargs["attachments"] = attachments
        await self.view.message.edit(view=self.view, **kwargs)


class StopTrackButton(discord.ui.Button):
    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        style: discord.ButtonStyle,
        row: int = None,
    ):
        super().__init__(
            style=style,
            emoji=emojis.STOP,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True)
        context = await self.cog.bot.get_context(interaction)
        if not await self.view.bot.is_owner(context.author):
            await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=context, description=_("You are not authorized to perform this action.")
                ),
                ephemeral=True,
            )
            return
        player = self.view.source.current_player
        if not player:
            await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=context,
                    description=_("There are no players available for this action. Try refreshing."),
                ),
                ephemeral=True,
            )
            return

        await player.stop(interaction.user)
        if notify_channel := await player.notify_channel():
            with contextlib.suppress(discord.HTTPException):
                await notify_channel.send(
                    embed=await self.cog.pylav.construct_embed(
                        title=_("My bot owner requested an action"),
                        description=_(
                            "My owner has requested that I wipe the queue for this server and stop playback. Sorry for any inconvenience caused."
                        ),
                    )
                )

        await self.view.prepare()
        kwargs = await self.view.get_page(self.view.current_page)
        attachments = []
        if "file" in kwargs:
            attachments = [kwargs.pop("file")]
        elif "files" in kwargs:
            attachments = kwargs.pop("files")
        if attachments:
            kwargs["attachments"] = attachments
        await self.view.message.edit(view=self.view, **kwargs)


class DisconnectAllButton(discord.ui.Button):
    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        disconnect_type: Literal["all", "inactive"],
        style: discord.ButtonStyle,
        row: int = None,
    ):
        super().__init__(
            style=style,
            emoji=emojis.POWER,
            row=row,
        )

        self.disconnect_type = disconnect_type
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True)
        context = await self.cog.bot.get_context(interaction)
        if not await self.view.bot.is_owner(context.author):
            await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=context, description=_("You are not authorized to perform this action.")
                ),
                ephemeral=True,
            )
            return

        players = (
            self.cog.pylav.player_manager.connected_players
            if self.disconnect_type == "all"
            else self.cog.pylav.player_manager.not_playing_players
        )
        if not players:
            await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=context,
                    description=_("There are no players available for this action. Try refreshing."),
                ),
                ephemeral=True,
            )
            return
        for player in players:
            if notify_channel := await player.notify_channel():
                with contextlib.suppress(discord.HTTPException):
                    await notify_channel.send(
                        embed=await self.cog.pylav.construct_embed(
                            title=_("My bot owner requested an action"),
                            description=_(
                                "My bot owner has requested that I disconnect from all servers. "
                                "I have saved my current position, so connect me to a voice channel to resume playback. "
                                "Sorry for any inconvenience caused."
                            ),
                        )
                    )
            await player.disconnect(requester=context.author)

        await self.view.prepare()
        kwargs = await self.view.get_page(self.view.current_page)
        attachments = []
        if "file" in kwargs:
            attachments = [kwargs.pop("file")]
        elif "files" in kwargs:
            attachments = kwargs.pop("files")
        if attachments:
            kwargs["attachments"] = attachments
        await self.view.message.edit(view=self.view, **kwargs)



from __future__ import annotations

from pylav.players.filters.misc import FilterMixin


class Rotation(FilterMixin):
    __slots__ = ("_hertz", "_default")

    def __init__(self, hertz: float | None = None) -> None:
        super().__init__()
        self.hertz = hertz

    def to_dict(self) -> dict[str, float | bool | None]:
        return {
            "hertz": self.hertz,
        }

    @classmethod
    def from_dict(cls, data: dict[str, float | bool | None]) -> Rotation:
        return cls(hertz=data["hertz"])

    def __repr__(self) -> str:
        return f"<Rotation: hertz={self.hertz}>"

    @property
    def hertz(self) -> float | None:
        return self._hertz

    @hertz.setter
    def hertz(self, v: float | None) -> None:
        self._hertz = v

    @classmethod
    def default(cls) -> Rotation:
        return cls()

    def get(self) -> dict[str, float]:
        if self.off:
            return {}
        response = {}
        if self.hertz is not None:
            response["rotationHz"] = self.hertz
        return response

    def reset(self) -> None:
        self.hertz = None



from __future__ import annotations

from typing import TYPE_CHECKING

import discord

from pylav.events.base import PyLavEvent

if TYPE_CHECKING:
    from pylav.nodes.api.responses.websocket import Closed
    from pylav.nodes.node import Node
    from pylav.players.player import Player


class NodeDisconnectedEvent(PyLavEvent):
    """This event is dispatched when a node disconnects and becomes unavailable.

    Event can be listened to by adding a listener with the name `pylav_node_disconnected_event`.

    Attributes
    ----------
    node: :class:`Node`
        The node that disconnected.
    code: :class:`int`
        The close code.
    reason: :class:`str`
        The close reason.

    """

    __slots__ = ("node", "code", "reason")

    def __init__(self, node: Node, code: int, reason: str) -> None:
        self.node = node
        self.code = code
        self.reason = reason


class NodeConnectedEvent(PyLavEvent):
    """This event is dispatched when PyLav successfully connects to a node.

    Event can be listened to by adding a listener with the name `pylav_node_connected_event`.

    Attributes
    ----------
    node: :class:`Node`
        The node that connected.
    """

    __slots__ = ("node",)

    def __init__(self, node: Node) -> None:
        self.node = node


class NodeChangedEvent(PyLavEvent):
    """This event is dispatched when a player changes to another node.
    Keep in mind this event can be dispatched multiple times if a node
    disconnects and the load balancer moves players to a new node.

    Event can be listened to by adding a listener with the name `pylav_node_changed_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that changed nodes.
    old_node: :class:`Node`
        The node the player was on before the change.
    new_node: :class:`Node`
        The node the player is on after the change.
    """

    __slots__ = ("player", "old_node", "new_node")

    def __init__(self, player: Player, old_node: Node, new_node: Node) -> None:
        self.player = player
        self.old_node = old_node
        self.new_node = new_node


class WebSocketClosedEvent(PyLavEvent):
    """This event is dispatched when an audio websocket to Discord
    is closed. This can happen for various reasons like an
    expired voice server update.

    Event can be listened to by adding a listener with the name `pylav_websocket_closed_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player whose websocket was closed.
    node: :class:`Node`
        The node the player is on.
    channel: :class:`discord.channel.VocalGuildChannel`
        The channel the player is in.
    code: :class:`int`
        The close code.
    reason: :class:`str`
        The close reason.
    by_remote: :class:`bool`
        Whether the close was initiated by Discord or not.
    event: :class:`Closed`
        The received event object.
    """

    __slots__ = ("player", "node", "channel", "code", "reason", "by_remote", "event")

    def __init__(
        self,
        player: Player,
        node: Node,
        channel: discord.channel.VocalGuildChannel,
        event_object: Closed,
    ) -> None:
        self.player = player
        self.code = event_object.code
        self.reason = event_object.reason
        self.by_remote = event_object.byRemote
        self.node = node
        self.channel = channel
        self.event = event_object



from __future__ import annotations

import math
import platform
import sys

import psutil

from pylav._internals.functions import get_true_path
from pylav.type_hints.dict_typing import JSON_DICT_TYPE


def get_max_allocation_size(executable: str) -> tuple[int, bool]:
    """Returns the maximum heap size allowed for the given executable."""
    if platform.architecture(executable)[0] == "64bit":
        max_heap_allowed = psutil.virtual_memory().total
        thinks_is_64_bit = True
    else:
        max_heap_allowed = min(4 * 1024**3, psutil.virtual_memory().total)
        thinks_is_64_bit = False
    return max_heap_allowed, thinks_is_64_bit


def _calculate_ram(max_allocation: int) -> tuple[str, str, int, int]:
    min_ram_int = 64 * 1024**2
    _min_max = 512 * 1024**2
    max_ram_allowed = max(max_allocation * 0.5, _min_max)
    max_ram_int = max(min_ram_int, max_ram_allowed)
    size_name = ("", "K", "M", "G", "T")
    i = int(math.floor(math.log(min_ram_int, 1024)))
    p = math.pow(1024, i)
    s = int(min_ram_int // p)
    min_ram = f"{s}{size_name[i]}"

    i = int(math.floor(math.log(max_ram_int, 1024)))
    p = math.pow(1024, i)
    s = int(max_ram_int // p)
    max_ram = f"{s}{size_name[i]}"

    return min_ram, max_ram, min_ram_int, max_ram_int


def get_jar_ram_defaults() -> tuple[str, str, int, int]:
    """Returns the default ram for the jar"""
    # We don't know the java executable at this stage - not worth the extra work required here
    max_allocation, __ = get_max_allocation_size(sys.executable)
    min_ram, max_ram, min_ram_int, max_ram_int = _calculate_ram(max_allocation)
    return min_ram, max_ram, min_ram_int, max_ram_int


def get_jar_ram_actual(executable: str) -> tuple[str, str, int, int]:
    """Returns the actual ram for the jar"""
    if not executable:
        from pylav.constants.config import JAVA_EXECUTABLE

        executable = JAVA_EXECUTABLE
    executable = get_true_path(executable, sys.executable)
    max_allocation, __ = get_max_allocation_size(executable)
    min_ram, max_ram, min_ram_int, max_ram_int = _calculate_ram(max_allocation)
    return min_ram, max_ram, min_ram_int, max_ram_int


def convert_function(key: str) -> str:
    """Converts a key to a valid key."""
    return key.replace("_", "-")


def change_dict_naming_convention(data: JSON_DICT_TYPE) -> JSON_DICT_TYPE:
    """Changes the naming convention of a dict."""
    new = {}
    for k, v in data.items():
        new_v = v
        if isinstance(v, dict):
            new_v = change_dict_naming_convention(v)
        elif isinstance(v, list):
            new_v = []
            for x in v:
                if isinstance(x, dict):
                    new_v.append(change_dict_naming_convention(x))
                else:
                    new_v.append(x)
        new[convert_function(k)] = new_v
    return new



from __future__ import annotations

import asyncio
import contextlib
import dataclasses
import datetime
import functools
import logging
from typing import TYPE_CHECKING, Any
from uuid import uuid4

import aiohttp
from aiohttp.helpers import sentinel  # noqa:
from apscheduler.jobstores.base import JobLookupError
from dacite import from_dict
from expiringdict import ExpiringDict
from multidict import CIMultiDictProxy
from packaging.version import Version, parse
from yarl import URL

from pylav.compat import json
from pylav.constants.builtin_nodes import BUNDLED_NODES_IDS_HOST_MAPPING, PYLAV_NODES
from pylav.constants.coordinates import REGION_TO_COUNTRY_COORDINATE_MAPPING
from pylav.constants.node import GOOD_RESPONSE_RANGE, MAX_SUPPORTED_API_MAJOR_VERSION
from pylav.constants.node_features import SUPPORTED_FEATURES, SUPPORTED_SOURCES
from pylav.constants.regex import SEMANTIC_VERSIONING
from pylav.events.api import LavalinkLoadtracksEvent
from pylav.events.base import PyLavEvent
from pylav.exceptions.request import HTTPException, UnauthorizedException
from pylav.helpers.time import get_now_utc
from pylav.logging import getLogger
from pylav.nodes.api.responses import rest_api
from pylav.nodes.api.responses import websocket as websocket_responses
from pylav.nodes.api.responses.errors import LavalinkError
from pylav.nodes.api.responses.route_planner import Status as RoutePlannerStart
from pylav.nodes.api.responses.track import Track
from pylav.nodes.utils import EMPTY_RESPONSE, Stats
from pylav.nodes.websocket import WebSocket
from pylav.players.filters import (
    ChannelMix,
    Distortion,
    Echo,
    Equalizer,
    Karaoke,
    LowPass,
    Rotation,
    Timescale,
    Tremolo,
    Vibrato,
    Volume,
)
from pylav.players.query.obj import Query
from pylav.storage.models.node import mocked as node_mocked
from pylav.storage.models.node import real as node_real
from pylav.type_hints.dict_typing import JSON_DICT_TYPE
from pylav.utils.location import distance

if TYPE_CHECKING:
    from pylav.nodes.manager import NodeManager
    from pylav.players.player import Player


class Node:
    """Represents a Node connection with Lavalink.

    Note
    ----
    Nodes are **NOT** meant to be added manually, but rather with :func:`Client.add_node`.
    """

    __slots__ = (
        "_query_cls",
        "_manager",
        "_session",
        "_temporary",
        "_host",
        "_port",
        "_password",
        "_name",
        "_ssl",
        "_config",
        "_managed",
        "_region",
        "_extras",
        "_stats",
        "_disabled_sources",
        "_identifier",
        "_resume_timeout",
        "_reconnect_attempts",
        "_search_only",
        "_capabilities",
        "_coordinates",
        "_down_votes",
        "_ready",
        "_ws",
        "_version",
        "_api_version",
        "_logger",
        "_filters",
        "__cli_flags",
    )

    def __init__(
        self,
        manager: NodeManager,
        host: str,
        password: str,
        resume_timeout: int,
        port: int | None = None,
        name: str | None = None,
        reconnect_attempts: int = -1,
        ssl: bool = False,
        search_only: bool = False,
        unique_identifier: int = None,
        disabled_sources: list[str] = None,
        managed: bool = False,
        extras: dict = None,
        temporary: bool = False,
    ) -> None:
        self._query_cls: Query = Query  # type: ignore
        self._version: Version | None = None
        self._api_version: int | None = None
        self._manager = manager
        self._session = aiohttp.ClientSession(timeout=aiohttp.ClientTimeout(total=120), json_serialize=json.dumps)
        self._temporary = temporary
        if not temporary:
            # noinspection PyProtectedMember
            self._config: node_real.Node = self._manager._client.node_db_manager.get_node_config(unique_identifier)
        else:
            self._config: node_mocked.NodeMock | None = None
        if unique_identifier is None:
            unique_identifier = str(uuid4())
        self._managed = managed
        self._region = None
        self._host = host
        self._name = name or f"{self._region}-{self._host}-{unique_identifier}"
        self._extras = extras or {}
        self._disabled_sources = set(disabled_sources or [])

        self._logger = getLogger(f"PyLav.Node-{self._name}")

        if self._manager.get_node_by_id(unique_identifier) is not None:
            raise ValueError(f"A Node with identifier:{unique_identifier} already exists")
        self._identifier = unique_identifier
        self._ssl = ssl
        if port is None:
            self._port = 443 if self._ssl else 80
        else:
            self._port = port
        self._password = password

        self._resume_timeout = resume_timeout
        self._reconnect_attempts = reconnect_attempts
        self._search_only = search_only
        self._capabilities: set[str] = set()
        self._filters: set[str] = set()
        self._coordinates = (0, 0)
        self._down_votes = ExpiringDict(max_len=float("inf"), max_age_seconds=600)  # type: ignore
        self.__cli_flags = getattr("manager._client.bot", "_cli_flags", None)

        self._stats = None

        self._ready = asyncio.Event()
        self._ws = WebSocket(
            node=self,
            host=self.host,
            port=self.port,
            password=self.password,
            resume_timeout=self.resume_timeout,
            reconnect_attempts=self.reconnect_attempts,
            ssl=self.ssl,
        )
        self._manager.client.scheduler.add_job(
            self.node_monitor_task,
            trigger="interval",
            seconds=15,
            max_instances=1,
            id=f"{self.identifier}-{self._manager.client.bot.user.id}-node_monitor_task",
            replace_existing=True,
            coalesce=True,
            next_run_time=get_now_utc() + datetime.timedelta(seconds=15),
        )

    @property
    def trace(self) -> bool:
        """Returns whether the node is being traced or not."""
        if getLogger("PyLav").getEffectiveLevel() < logging.INFO:
            return True
        return self.__cli_flags.logging_level < logging.INFO if self.__cli_flags else False

    async def _unhealthy(self) -> None:
        del self.down_votes
        if self._ws is not None:
            await self.websocket.manual_closure(
                managed_node=self.identifier == self.node_manager.client.bot.user.id and self.websocket is not None
            )
        if self.identifier == self.node_manager.client.bot.user.id:
            await self.node_manager.client.managed_node_controller.restart()
            with contextlib.suppress(Exception):
                await self.close()

    async def node_monitor_task(self) -> None:
        """Monitors the node for health and stability."""
        with contextlib.suppress(
            asyncio.exceptions.CancelledError,
        ):
            if self.websocket and self.websocket._connecting is True:
                return

            try:
                await self.websocket.ping()
                self._logger.trace("Healthy")
            except Exception:  # noqa
                # noinspection PyProtectedMember
                if self.websocket._connecting is True:
                    return
                self._logger.warning("Unhealthy - Triggering a state reset")
                await self._unhealthy()

            playing_players = len(self.playing_players)
            if playing_players == 0:
                return
            if (self.down_votes / playing_players) >= 0.5:
                # noinspection PyProtectedMember
                if self.websocket._connecting is True:
                    return
                await self._unhealthy()

    @property
    def version(self) -> Version | None:
        """The version of the node."""
        return self._version

    @property
    def api_version(self) -> int | None:
        """The API version of the node."""
        return self._api_version

    @property
    def socket_protocol(self) -> str:
        """The protocol used for the socket connection"""
        return "wss" if self._ssl else "ws"

    @property
    def is_ready(self) -> bool:
        """Whether the node is ready or not."""
        return self._ready.is_set() and self.websocket.connected

    @property
    def coordinates(self) -> tuple[int, int]:
        """The coordinates of the node.

        Returns
        -------
        :class:`tuple`
            The coordinates of the node.
        """
        return self._coordinates

    @property
    def managed(self) -> bool:
        """Whether the node is managed or not."""
        return self._managed

    @property
    def config(self) -> node_mocked.NodeMock | node_real.Node:
        """The config of the node."""
        return self._config

    @property
    def identifier(self) -> int:
        """
        The identifier of the :class:`Node`.
        """
        return self._identifier

    @property
    def search_only(self) -> bool:
        """Whether the node is search only or not."""
        return self._search_only

    @property
    def session(self) -> aiohttp.ClientSession:
        """The aiohttp session of the node"""
        return self._session

    @property
    def websocket(self) -> WebSocket:
        """The websocket of the node"""
        return self._ws

    @property
    def node_manager(self) -> NodeManager:
        """The :class:`NodeManager` this node belongs to"""
        return self._manager

    @property
    def port(self) -> int:
        """The port of the node"""
        return self._port

    @property
    def ssl(self) -> bool:
        """Whether the node is using a ssl connection"""
        return self._ssl

    @property
    def connection_protocol(self) -> str:
        """The protocol used for the connection"""
        return "https" if self.ssl else "http"

    @property
    def host(self) -> str:
        """The host of the node"""
        return self._host

    @property
    def region(self) -> str:
        """The region of the node"""
        return self._region

    @property
    def name(self) -> str:
        """The name of the node"""
        return self._name

    @property
    def password(self) -> str:
        """The password of the node"""
        return self._password

    @property
    def resume_timeout(self) -> int:
        """The timeout to use for resuming"""
        return self._resume_timeout

    @property
    def reconnect_attempts(self) -> int:
        """The number of attempts to reconnect to the node"""
        return self._reconnect_attempts

    @property
    def stats(self) -> Stats:
        """The stats of the node"""
        return self._stats

    @stats.setter
    def stats(self, value: Stats) -> None:
        """Sets the stats of the node"""
        if not isinstance(value, Stats):
            raise TypeError("stats must be of type Stats")
        self._stats = value

    @property
    def available(self) -> bool:
        """Returns whether the node is available for requests"""
        return self._ws.connected if self._ws else False

    @property
    def _original_players(self) -> list[Player]:
        """Returns a list of players that were assigned to this node, but were moved due to failover etc"""
        # noinspection PyProtectedMember
        return [p for p in self._manager.client.player_manager.players.values() if p._original_node == self]

    @property
    def players(self) -> list[Player]:
        """Returns a list of all players on this node"""
        return [p for p in self._manager.client.player_manager.players.values() if p.node == self]

    @property
    def playing_players(self) -> list[Player]:
        """Returns a list of all players on this node that are playing"""
        return [p for p in self.players if p.is_active]

    @property
    def connected_players(self) -> list[Player]:
        """Returns a list of all players on this node that are connected"""
        return [p for p in self.players if p.is_connected]

    @property
    def server_connected_players(self) -> int:
        """Returns the number of players on this node that are connected"""
        return self.stats.players if self.stats else self.connected_count

    @property
    def server_playing_players(self) -> int:
        """Returns the number of players on this node that are playing"""
        return self.stats.playing_players if self.stats else self.playing_count

    @property
    def count(self) -> int:
        """Returns the number of players on this node"""
        return len(self.players)

    @property
    def playing_count(self) -> int:
        """Returns the number of players on this node that are playing"""
        return len(self.playing_players)

    @property
    def connected_count(self) -> int:
        """Returns the number of players on this node that are connected"""
        return len(self.connected_players)

    @property
    def penalty(self) -> float:
        """Returns the load-balancing penalty for this node"""
        if not self.available or not self.stats:
            return float("inf")
        return self.stats.penalty.total

    @property
    def session_id(self) -> str:
        """Returns the session id of the node"""
        return self.websocket.session_id if self.websocket else ""

    def down_vote(self, player: Player) -> int:
        """Adds a down vote for this node"""
        if not player.is_active:
            return -1
        self._down_votes[player.guild.id] = 1
        return self.down_votes

    def down_unvote(self, player: Player) -> int:
        """Removes a down vote for this node"""
        if not player.is_active:
            return -1
        self._down_votes.pop(player.guild.id, None)
        return self.down_votes

    @property
    def down_votes(self) -> int:
        """Returns the down votes for this node"""
        return len(set(self._down_votes.keys()))

    def voted(self, player: Player) -> bool:
        """Returns whether a player has voted for this node"""
        return player.guild.id in self._down_votes

    @down_votes.deleter
    def down_votes(self):
        """Clears the down votes for this node"""
        self._down_votes.clear()

    @property
    def can_resume(self) -> bool:
        """Returns whether the node can be resumed"""
        return self._ws.can_resume

    async def penalty_with_region(self, region: str | None) -> float:
        """The penalty for the node, with the region added in"""
        if not region:
            return self.penalty
        return self.penalty + (1.1 ** (0.0025 * await self.region_distance(region)) * 500 - 500)

    def dispatch_event(self, event: PyLavEvent) -> None:
        """|coro|
        Dispatches the given event to all registered hooks.
        Parameters
        ----------
        event: :class:`Event`
            The event to dispatch to the hooks.
        """
        self.node_manager.client.dispatch_event(event)

    def __repr__(self) -> str:
        return (
            f"<Node id={self.identifier} name={self.name} session_id={self.session_id} "
            f"region={self.region} ssl={self.ssl} "
            f"search_only={self.search_only} connected={self.websocket.connected if self._ws else False} "
            f"votes={self.down_votes} "
            f"players={self.server_connected_players} playing={self.server_playing_players}>"
        )

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Node):
            return functools.reduce(
                lambda x, y: x and y,
                map(
                    lambda p, q: p == q,
                    [self.identifier, self.websocket.connected, self.name, self.session_id],
                    [other.identifier, self.websocket.connected, self.name, other.session_id],
                ),
                True,
            )
        elif isinstance(other, (node_real.Node, node_mocked.NodeMock)):
            return self.identifier == other.id
        return NotImplemented

    def __ne__(self, other) -> bool:
        if isinstance(other, Node):
            return self.identifier != other.identifier
        elif isinstance(other, (node_real.Node, node_mocked.NodeMock)):
            return self.identifier != other.id
        return NotImplemented

    def __hash__(self) -> int:
        return hash((self.host, self.port))

    @staticmethod
    def parse_loadtrack_response(
        data: JSON_DICT_TYPE,
    ) -> (
        rest_api.ErrorResponse
        | rest_api.EmptyResponse
        | rest_api.PlaylistResponse
        | rest_api.SearchResponse
        | rest_api.TrackResponse
    ):
        """Parses the loadtrack response.
         Parameters
         ----------
         data: LoadTracksResponseT
             The data to parse.
        Returns
         -------
         LavalinkLoadTrackObjects
             Lavalink LoadTrack Response Object
        """

        match data["loadType"]:
            case "error":
                return from_dict(data_class=rest_api.ErrorResponse, data=data)
            case "empty":
                return from_dict(data_class=rest_api.EmptyResponse, data=data)
            case "playlist":
                return from_dict(data_class=rest_api.PlaylistResponse, data=data)
            case "track":
                return from_dict(data_class=rest_api.TrackResponse, data=data)
            case "search":
                return from_dict(data_class=rest_api.SearchResponse, data=data)

    async def get_unsupported_features(self) -> set[str]:
        """|coro|
        Returns the unsupported features of the target node.
        """
        await self.update_features()
        return SUPPORTED_SOURCES.union(SUPPORTED_FEATURES) - self._capabilities

    async def update_features(self) -> set[str]:
        """|coro|
        Updates the features of the target node.
        """
        info = await self.fetch_info()
        self._capabilities.clear()
        self._filters.clear()
        for source in info.sourceManagers:
            self._capabilities.add(source)
        for filterName in info.filters:
            self._filters.add(filterName)
        for plugin in info.plugins:
            match plugin.name:
                case "SponsorBlock-Plugin":
                    self._capabilities.add("sponsorblock")
        if self.identifier in PYLAV_NODES:
            self._capabilities.discard("http")
            self._capabilities.discard("local")
        # If not setup says these should be disabled remove them to trick the node to think they are disabled
        if self._capabilities:
            self._capabilities.difference_update(self._disabled_sources)
        return self._capabilities.copy()

    def has_source(self, source: str) -> bool:
        """
        Checks if the target node has the specified source.

        Parameters
        ----------
        source: :class:`str`
            The source to check.

        Returns
        -------
        :class:`bool`
            True if the target node has the specified source, False otherwise.
        """
        return source.lower() in self.sources

    has_capability = has_source

    def has_filter(self, filter_name: str) -> bool:
        """
        Checks if the target node has the specified filter.

        Parameters
        ----------
        filter_name: :class:`str`
            The filter to check.

        Returns
        -------
        :class:`bool`
            True if the target node has the specified filter, False otherwise.
        """
        return filter_name in self._filters

    async def update_disabled_sources(self, sources: set[str]) -> None:
        """
        Updates the disabled sources.

        Returns
        -------
        :class:`None`
        """
        if self.managed or self.identifier in BUNDLED_NODES_IDS_HOST_MAPPING or self.identifier == 31415:
            return
        unsupported = await self.get_unsupported_features()
        currently_disabled = set(await self.config.fetch_disabled_sources())
        unsupported = list(unsupported.union(currently_disabled).union(sources))
        await self.config.update_disabled_sources(unsupported)
        self._disabled_sources = unsupported

    @property
    def capabilities(self) -> set:
        """
        Returns the capabilities of the target node.

        Returns
        -------
        :class:`set`
            The capabilities of the target node.
        """
        return self._capabilities.copy()

    @property
    def disabled_sources(self) -> set:
        """
        Returns the disabled sources of the target node.

        Returns
        -------
        :class:`set`
            The disabled sources of the target node.
        """
        return self._disabled_sources.copy()

    @property
    def sources(self) -> set:
        """
        Returns the sources of the target node.

        Returns
        -------
        :class:`set`
            The sources of the target node.
        """
        return self._capabilities.copy()

    @property
    def supports_spotify(self) -> bool:
        """
        Checks if the target node supports Spotify.

        Returns
        -------
        :class:`bool`
            True if the target node supports Spotify, False otherwise.
        """
        return self.has_source("spotify")

    @property
    def supports_apple_music(self) -> bool:
        """
        Checks if the target node supports Apple Music.

        Returns
        -------
        :class:`bool`
            True if the target node supports Apple Music, False otherwise.
        """
        return self.has_source("applemusic")

    @property
    def supports_getyarn(self) -> bool:
        """
        Checks if the target node supports GetYarn.

        Returns
        -------
        :class:`bool`
            True if the target node supports GetYarn, False otherwise.
        """
        return self.has_source("getyarn.io")

    @property
    def supports_soundgasm(self) -> bool:
        """
        Checks if the target node supports Soundgasm.

        Returns
        -------
        :class:`bool`
            True if the target node supports Soundgasm, False otherwise.
        """
        return self.has_source("soundgasm")

    @property
    def supports_clypit(self) -> bool:
        """
        Checks if the target node supports ClypIt.

        Returns
        -------
        :class:`bool`
            True if the target node supports ClypIt, False otherwise.
        """
        return self.has_source("clypit")

    @property
    def supports_speak(self) -> bool:
        """
        Checks if the target node supports speak source.

        Returns
        -------
        :class:`bool`
            True if the target node supports speak, False otherwise.
        """
        return self.has_source("speak")

    @property
    def supports_tts(self) -> bool:
        """
        Checks if the target node supports Google Cloud TTS.

        Returns
        -------
        :class:`bool`
            True if the target node supports Google Cloud TTS, False otherwise.
        """
        return self.has_capability("gcloud-tts")

    @property
    def supports_pornhub(self) -> bool:
        """
        Checks if the target node supports PornHub.

        Returns
        -------
        :class:`bool`
            True if the target node supports PornHub, False otherwise.
        """
        return self.has_source("pornhub")

    @property
    def supports_reddit(self) -> bool:
        """
        Checks if the target node supports Reddit.

        Returns
        -------
        :class:`bool`
            True if the target node supports Reddit, False otherwise.
        """
        return self.has_source("reddit")

    @property
    def supports_ocremix(self) -> bool:
        """
        Checks if the target node supports OCRemix.

        Returns
        -------
        :class:`bool`
            True if the target node supports OCRemix, False otherwise.
        """
        return self.has_source("ocremix")

    @property
    def supports_mixcloud(self) -> bool:
        """
        Checks if the target node supports Mixcloud.

        Returns
        -------
        :class:`bool`
            True if the target node supports Mixcloud, False otherwise.
        """
        return self.has_source("mixcloud")

    @property
    def supports_tiktok(self) -> bool:
        """
        Checks if the target node supports TikTok.

        Returns
        -------
        :class:`bool`
            True if the target node supports TikTok, False otherwise.
        """
        return self.has_source("tiktok")

    @property
    def supports_youtube(self) -> bool:
        """
        Checks if the target node supports YouTube.

        Returns
        -------
        :class:`bool`
            True if the target node supports YouTube, False otherwise.
        """
        return self.has_source("youtube")

    @property
    def supports_bandcamp(self) -> bool:
        """
        Checks if the target node supports Bandcamp.

        Returns
        -------
        :class:`bool`
            True if the target node supports Bandcamp, False otherwise.
        """
        return self.has_source("bandcamp")

    @property
    def supports_soundcloud(self) -> bool:
        """
        Checks if the target node supports SoundCloud.

        Returns
        -------
        :class:`bool`
            True if the target node supports SoundCloud, False otherwise.
        """
        return self.has_source("soundcloud")

    @property
    def supports_twitch(self) -> bool:
        """
        Checks if the target node supports Twitch.

        Returns
        -------
        :class:`bool`
            True if the target node supports Twitch, False otherwise.
        """
        return self.has_source("twitch")

    @property
    def supports_deezer(self) -> bool:
        """
        Checks if the target node supports Deezer.

        Returns
        -------
        :class:`bool`
            True if the target node supports Deezer, False otherwise.
        """
        return self.has_source("deezer")

    @property
    def supports_yandex_music(self) -> bool:
        """
        Checks if the target node supports Yandex Music.

        Returns
        -------
        :class:`bool`
            True if the target node supports Yandex Music, False otherwise.
        """
        return self.has_source("yandexmusic")

    @property
    def supports_vimeo(self) -> bool:
        """
        Checks if the target node supports Vimeo.

        Returns
        -------
        :class:`bool`
            True if the target node supports Vimeo, False otherwise.
        """
        return self.has_source("vimeo")

    @property
    def supports_http(self) -> bool:
        """
        Checks if the target node supports HTTP.

        Returns
        -------
        :class:`bool`
            True if the target node supports HTTP, False otherwise.
        """
        return self.has_source("http")

    @property
    def supports_local(self) -> bool:
        """
        Checks if the target node supports local files.

        Returns
        -------
        :class:`bool`
            True if the target node supports local files, False otherwise.
        """
        return self.has_source("local")

    @property
    def supports_sponsorblock(self) -> bool:
        """
        Checks if the target node supports SponsorBlock.

        Returns
        -------
        :class:`bool`
            True if the target node supports SponsorBlock, False otherwise.
        """
        return self.has_capability("sponsorblock")

    async def close(self) -> None:
        """
        Closes the target node.
        """
        if self.websocket is not None:
            await self.websocket.close()
        await self.session.close()
        with contextlib.suppress(JobLookupError):
            self.node_manager.client.scheduler.remove_job(
                job_id=f"{self.identifier}-{self._manager.client.bot.user.id}-node_monitor_task"
            )

    async def wait_until_ready(self, timeout: float | None = None):
        """Waits until the target node is ready."""
        await asyncio.wait_for(self._ready.wait(), timeout=timeout)

    async def region_distance(self, region: str) -> float:
        """
        Returns the numeric representation of the distance between the target node and the given region.

        Parameters
        ----------
        region : :class:`str`
            The region to get the distance to.

        Returns
        -------
        :class:`float`
            The numeric representation of the distance between the target node and the given region.
        """
        coordinates = REGION_TO_COUNTRY_COORDINATE_MAPPING.get(region)

        return distance(*self.coordinates, *coordinates) if (coordinates and self.coordinates) else float("inf")

    async def get_track_from_cache(
        self, query: Query, first: bool = False
    ) -> rest_api.PlaylistResponse | rest_api.SearchResponse | rest_api.TrackResponse | None:
        """Gets a query from the query cache."""
        response = await self.node_manager.client.query_cache_manager.fetch_query(query)
        if not response:
            return
        if tracks := await response.fetch_tracks():
            load_type = "playlist" if query.is_playlist or query.is_album else "search" if query.is_search else "track"
            try:
                if tracks and first:
                    tracks = [tracks[0]]
            except Exception:  # noqa
                tracks = []
                load_type = "error"

            data = {"loadType": load_type, "data": None}
            match data["loadType"]:
                case "playlist":
                    data["data"] = {
                        "info": {"name": await response.fetch_name() or "", "selectedTrack": 0},
                        "pluginInfo": await response.fetch_plugin_info(),
                        "tracks": tracks,
                    }
                case "search":
                    data["data"] = tracks
                case "track":
                    data["data"] = tracks[0]
                case "empty":
                    data["data"] = None
                case "error":
                    data["data"] = {"cause": "No tracks returned", "severity": "common", "message": "No tracks found"}
            response = self.parse_loadtrack_response(data)
            return response

    @property
    def base_url(self) -> URL:
        """Returns the base URL of the target node."""
        return URL(f"{self.connection_protocol}://{self.host}:{self.port}")

    @property
    def base_api_url(self) -> URL:
        """Returns the base API URL of the target node."""
        return self.base_url / f"v{self.api_version}"

    @property
    def base_ws_url(self) -> URL:
        """Returns the base WebSocket URL of the target node."""
        return self.base_api_url.with_scheme(self.socket_protocol)

    # ENDPOINTS
    def get_endpoint_websocket(self) -> URL:
        """Returns the WebSocket endpoint of the target node."""
        return self.base_ws_url / "websocket"

    def get_endpoint_info(self) -> URL:
        """Returns the info endpoint of the target node."""
        return self.base_api_url / "info"

    def get_endpoint_session_players(self) -> URL:
        """Returns the session players endpoint of the target node."""
        return self.get_endpoint_session() / "players"

    def get_endpoint_session_player_by_guild_id(self, guild_id: int) -> URL:
        """Returns the session player endpoint of the target node for the given guild ID."""
        return self.get_endpoint_session_players() / f"{guild_id}"

    def get_endpoint_session(self) -> URL:
        """Returns the session endpoint of the target node."""
        return self.base_api_url / "sessions" / self.session_id

    def get_endpoint_loadtracks(self) -> URL:
        """Returns the loadtracks endpoint of the target node."""
        return self.base_api_url / "loadtracks"

    def get_endpoint_decodetrack(self) -> URL:
        """Returns the decodetrack endpoint of the target node."""
        return self.base_api_url / "decodetrack"

    def get_endpoint_decodetracks(self) -> URL:
        """Returns the decodetracks endpoint of the target node."""
        return self.base_api_url / "decodetracks"

    def get_endpoint_stats(self) -> URL:
        """Returns the stats endpoint of the target node."""
        return self.base_api_url / "stats"

    def get_endpoint_routeplanner_status(self) -> URL:
        """Returns the routeplanner status endpoint of the target node."""
        return self.base_api_url / "routeplanner" / "status"

    def get_endpoint_routeplanner_free_address(self) -> URL:
        """Returns the routeplanner free address endpoint of the target node."""
        return self.base_api_url / "routeplanner" / "free" / "address"

    def get_endpoint_routeplanner_free_all(self) -> URL:
        """Returns the routeplanner free all endpoint of the target node."""
        return self.base_api_url / "routeplanner" / "free" / "all"

    def get_endpoint_session_player_sponsorblock_categories(self, guild_id: int) -> URL:
        """Returns the session player sponsorblock categories endpoint of the target node for the given guild ID."""
        return self.get_endpoint_session_player_by_guild_id(guild_id) / "sponsorblock" / "categories"

    def get_endpoint_version(self) -> URL:
        """Returns the version endpoint of the target node."""
        return self.base_url / "version"

    # REST API - Direct calls
    async def fetch_session_players(self) -> list[rest_api.LavalinkPlayer] | HTTPException:
        """|coro|
        Gets all players associated with the target node.

        Returns
        -------
        list[rest_api.LavalinkPlayer]
            A list of all players associated with the target node.
        """
        async with self._session.get(
            self.get_endpoint_session_players(),
            headers={
                "Authorization": self.password,
                "Client-Name": f"PyLav/{self.node_manager.client.lib_version}",
                "App-Id": self.node_manager.client._user_id,
            },
            params={"trace": "true" if self.trace else "false"},
        ) as res:
            if res.status in GOOD_RESPONSE_RANGE:
                return [from_dict(data_class=rest_api.LavalinkPlayer, data=t) for t in await res.json(loads=json.loads)]
            failure = from_dict(data_class=LavalinkError, data=await res.json(loads=json.loads))
            if res.status in [401, 403]:
                raise UnauthorizedException(failure)
            self._logger.trace("Failed to get session players: %d %s", failure.status, failure.message)
            return HTTPException(failure)

    async def fetch_session_player(self, guild_id: int) -> rest_api.LavalinkPlayer | HTTPException:
        """|coro|
        Gets the player associated with the target node and the given guild ID.
        """
        async with self._session.get(
            self.get_endpoint_session_player_by_guild_id(guild_id=guild_id),
            headers={
                "Authorization": self.password,
                "Client-Name": f"PyLav/{self.node_manager.client.lib_version}",
                "App-Id": self.node_manager.client._user_id,
            },
            params={"trace": "true" if self.trace else "false"},
        ) as res:
            if res.status in GOOD_RESPONSE_RANGE:
                return from_dict(data_class=rest_api.LavalinkPlayer, data=await res.json(loads=json.loads))
            failure = from_dict(data_class=LavalinkError, data=await res.json(loads=json.loads))
            if res.status in [401, 403]:
                raise UnauthorizedException(failure)
            self._logger.trace("Failed to get session player: %d %s", failure.status, failure.message)
            return HTTPException(failure)

    async def patch_session_player(
        self, guild_id: int, no_replace: bool = False, payload: JSON_DICT_TYPE = None
    ) -> rest_api.LavalinkPlayer | HTTPException:
        """|coro|
        Updates the player associated with the target node and the given guild ID.
        """
        async with self._session.patch(
            self.get_endpoint_session_player_by_guild_id(guild_id=guild_id),
            headers={
                "Authorization": self.password,
                "Client-Name": f"PyLav/{self.node_manager.client.lib_version}",
                "App-Id": self.node_manager.client._user_id,
            },
            params={"noReplace": "true" if no_replace else "false", "trace": "true" if self.trace else "false"},
            json=payload,
        ) as res:
            if res.status in GOOD_RESPONSE_RANGE:
                return from_dict(data_class=rest_api.LavalinkPlayer, data=await res.json(loads=json.loads))
            failure = from_dict(data_class=LavalinkError, data=await res.json(loads=json.loads))
            if res.status in [401, 403]:
                raise UnauthorizedException(failure)
            self._logger.trace("Failed to patch session player: %d %s", failure.status, failure.message)
            return HTTPException(failure)

    async def delete_session_player(self, guild_id: int) -> None | HTTPException:
        """|coro|
        Deletes the player associated with the target node and the given guild ID.
        """
        async with self._session.delete(
            self.get_endpoint_session_player_by_guild_id(guild_id=guild_id),
            headers={
                "Authorization": self.password,
                "Client-Name": f"PyLav/{self.node_manager.client.lib_version}",
                "App-Id": self.node_manager.client._user_id,
            },
            params={"trace": "true" if self.trace else "false"},
        ) as res:
            if res.status in GOOD_RESPONSE_RANGE or res.status in [404]:
                return
            response = await res.json(loads=json.loads)
            failure = from_dict(data_class=LavalinkError, data=response)
            if res.status in [401, 403]:
                raise UnauthorizedException(failure)
            self._logger.trace("Failed to delete session player: %d %s", failure.status, failure.message)
            return HTTPException(failure)

    async def get_session_player_sponsorblock_categories(self, guild_id: int) -> list[str] | HTTPException:
        """|coro|
        Gets the sponsorblock categories for the player associated with the target node and the given guild ID.
        """
        async with self._session.get(
            self.get_endpoint_session_player_sponsorblock_categories(guild_id=guild_id),
            headers={
                "Authorization": self.password,
                "Client-Name": f"PyLav/{self.node_manager.client.lib_version}",
                "App-Id": self.node_manager.client._user_id,
            },
        ) as res:
            if res.status in GOOD_RESPONSE_RANGE:
                return await res.json(loads=json.loads)
            failure = from_dict(data_class=LavalinkError, data=await res.json(loads=json.loads))
            if res.status in [401, 403]:
                raise UnauthorizedException(failure)
            self._logger.trace(
                "Failed to get session player sponsorblock categories: %d %s", failure.status, failure.message
            )
            return HTTPException(failure)

    async def put_session_player_sponsorblock_categories(
        self, guild_id: int, categories: list[str]
    ) -> None | HTTPException:
        """|coro|
        Sets the sponsorblock categories for the player associated with the target node and the given guild ID.
        """
        async with self._session.put(
            self.get_endpoint_session_player_sponsorblock_categories(guild_id=guild_id),
            headers={
                "Authorization": self.password,
                "Client-Name": f"PyLav/{self.node_manager.client.lib_version}",
                "App-Id": self.node_manager.client._user_id,
            },
            json=categories,
        ) as res:
            if res.status in GOOD_RESPONSE_RANGE:
                return
            failure = from_dict(data_class=LavalinkError, data=await res.json(loads=json.loads))
            if res.status in [401, 403]:
                raise UnauthorizedException(failure)
            self._logger.trace(
                "Failed to put session player sponsorblock categories: %d %s", failure.status, failure.message
            )
            return HTTPException(failure)

    async def delete_session_player_sponsorblock_categories(self, guild_id: int) -> None | HTTPException:
        """|coro|
        Deletes the sponsorblock categories for the player associated with the target node and the given guild ID.
        """
        async with self._session.delete(
            self.get_endpoint_session_player_sponsorblock_categories(guild_id=guild_id),
            headers={
                "Authorization": self.password,
                "Client-Name": f"PyLav/{self.node_manager.client.lib_version}",
                "App-Id": self.node_manager.client._user_id,
            },
        ) as res:
            if res.status in GOOD_RESPONSE_RANGE:
                return
            failure = from_dict(data_class=LavalinkError, data=await res.json(loads=json.loads))
            if res.status in [401, 403]:
                raise UnauthorizedException(failure)
            self._logger.trace(
                "Failed to delete session player sponsorblock categories: %d %s", failure.status, failure.message
            )
            return HTTPException(failure)

    async def patch_session(self, payload: JSON_DICT_TYPE) -> None | HTTPException:
        """|coro|
        Patches the session associated with the target node.
        """
        async with self._session.patch(
            self.get_endpoint_session(),
            headers={
                "Authorization": self.password,
                "Client-Name": f"PyLav/{self.node_manager.client.lib_version}",
                "App-Id": self.node_manager.client._user_id,
            },
            json=payload,
            params={"trace": "true" if self.trace else "false"},
        ) as res:
            if res.status in GOOD_RESPONSE_RANGE:
                return
            failure = from_dict(data_class=LavalinkError, data=await res.json(loads=json.loads))
            if res.status in [401, 403]:
                raise UnauthorizedException(failure)
            self._logger.trace("Failed to delete session player: %d %s", failure.status, failure.message)
            return HTTPException(failure)

    async def fetch_loadtracks(self, query: Query) -> rest_api.LoadTrackResponses:
        """|coro|
        Fetches the loadtracks response from the target node.
        """
        if not self.available or not self.has_source(query.requires_capability):
            return dataclasses.replace(EMPTY_RESPONSE)

        async with self._session.get(
            self.get_endpoint_loadtracks(),
            headers={
                "Authorization": self.password,
                "Client-Name": f"PyLav/{self.node_manager.client.lib_version}",
                "App-Id": self.node_manager.client._user_id,
            },
            params={"identifier": query.query_identifier, "trace": "true" if self.trace else "false"},
        ) as res:
            if res.status in GOOD_RESPONSE_RANGE:
                result = await res.json(loads=json.loads)
                response = self.parse_loadtrack_response(result)
                asyncio.create_task(self.node_manager.client.query_cache_manager.add_query(query, response))
                self._manager.client.dispatch_event(LavalinkLoadtracksEvent(node=self, response=response))
                return response
            failure = from_dict(data_class=LavalinkError, data=await res.json(loads=json.loads))
            if res.status in [401, 403]:
                raise UnauthorizedException(failure)
            self._logger.trace("Failed to load track: %d %s", failure.status, failure.message)
            return HTTPException(failure)

    async def fetch_decodetrack(
        self, encoded_track: str, timeout: aiohttp.ClientTimeout | object = sentinel, raise_on_failure: bool = True
    ) -> Track | HTTPException:
        """|coro|
        Fetches the decodetrack response from the target node.
        """
        async with self._manager._client.cached_session.get(
            self.get_endpoint_decodetrack(),
            headers={
                "Authorization": self.password,
                "Client-Name": f"PyLav/{self.node_manager.client.lib_version}",
                "App-Id": self.node_manager.client._user_id,
            },
            params={"encodedTrack": encoded_track, "trace": "true" if self.trace else "false"},
            timeout=timeout,
        ) as res:
            if res.status in GOOD_RESPONSE_RANGE:
                return from_dict(data_class=Track, data=await res.json(loads=json.loads))
            failure = from_dict(data_class=LavalinkError, data=await res.json(loads=json.loads))
            if res.status in [401, 403]:
                raise UnauthorizedException(failure)
            self._logger.trace("Failed to decode track: %d %s", failure.status, failure.message)
            if raise_on_failure:
                raise HTTPException(failure)
            return HTTPException(failure)

    async def post_decodetracks(
        self, encoded_tracks: list[str], raise_on_failure: bool = False
    ) -> list[Track] | HTTPException:
        """|coro|
        Posts the decodetracks response from the target node.
        """
        async with self._manager._client.cached_session.post(
            self.get_endpoint_decodetracks(),
            headers={
                "Authorization": self.password,
                "Client-Name": f"PyLav/{self.node_manager.client.lib_version}",
                "App-Id": self.node_manager.client._user_id,
            },
            json=encoded_tracks,
            params={"trace": "true" if self.trace else "false"},
        ) as res:
            if res.status in GOOD_RESPONSE_RANGE:
                return [from_dict(data_class=Track, data=t) for t in await res.json(loads=json.loads)]
            failure = from_dict(data_class=LavalinkError, data=await res.json(loads=json.loads))
            if res.status in [401, 403]:
                raise UnauthorizedException(failure)
            self._logger.trace("Failed to decode tracks: %d %s", failure.status, failure.message)
            if raise_on_failure:
                raise HTTPException(failure)
            return HTTPException(failure)

    async def fetch_info(self, raise_on_error: bool = False) -> rest_api.LavalinkInfo | HTTPException:
        """|coro|
        Fetches the info response from the target node.
        """
        async with self._session.get(
            self.get_endpoint_info(),
            headers={
                "Authorization": self.password,
                "Client-Name": f"PyLav/{self.node_manager.client.lib_version}",
                "App-Id": self.node_manager.client._user_id,
            },
            params={"trace": "true" if self.trace else "false"},
        ) as res:
            if res.status in GOOD_RESPONSE_RANGE:
                return from_dict(data_class=rest_api.LavalinkInfo, data=await res.json(loads=json.loads))
            failure = from_dict(data_class=LavalinkError, data=await res.json(loads=json.loads))
            if res.status in [401, 403]:
                if raise_on_error:
                    raise UnauthorizedException(failure)
                return UnauthorizedException(failure)
            self._logger.trace("Failed to get info: %d %s", failure.status, failure.message)
            if raise_on_error:
                raise HTTPException(failure)
            return HTTPException(failure)

    def _process_version_from_headers(self, headers: CIMultiDictProxy[str]) -> Version:
        self._api_version = int(headers.get("Lavalink-API-Version") or MAX_SUPPORTED_API_MAJOR_VERSION)
        version_str = f"{self._api_version}.9999.9999"
        return Version(version_str)

    async def fetch_stats(self, raise_on_error: bool = False) -> websocket_responses.Stats | HTTPException:
        """|coro|
        Fetches the stats response from the target node.
        """
        async with self._session.get(
            self.get_endpoint_stats(),
            headers={
                "Authorization": self.password,
                "Client-Name": f"PyLav/{self.node_manager.client.lib_version}",
                "App-Id": self.node_manager.client._user_id,
            },
            params={"trace": "true" if self.trace else "false"},
        ) as res:
            if res.status in GOOD_RESPONSE_RANGE:
                return from_dict(data_class=rest_api.Stats, data=await res.json(loads=json.loads))
            failure = from_dict(data_class=LavalinkError, data=await res.json(loads=json.loads))
            if res.status in [401, 403]:
                if raise_on_error:
                    raise UnauthorizedException(failure)
                return HTTPException(failure)
            self._logger.trace("Failed to get stats: %d %s", failure.status, failure.message)
            if raise_on_error:
                raise HTTPException(failure)
            return HTTPException(failure)

    async def fetch_version(self, raise_on_error: bool = False) -> Version | HTTPException:
        """|coro|
        Fetches the version response from the target node.
        """
        async with self._session.get(
            self.get_endpoint_version(),
            headers={
                "Authorization": self.password,
                "Content-Type": "text/plain",
                "Client-Name": f"PyLav/{self.node_manager.client.lib_version}",
                "App-Id": self.node_manager.client._user_id,
            },
            params={"trace": "true" if self.trace else "false"},
        ) as res:
            if res.status in GOOD_RESPONSE_RANGE:
                text = await res.text()
                version_from_header = self._process_version_from_headers(res.headers)
                return parse(text) if SEMANTIC_VERSIONING.match(text) else version_from_header
            failure = from_dict(data_class=LavalinkError, data=await res.json(loads=json.loads))
            if res.status in [401, 403]:
                if raise_on_error:
                    raise UnauthorizedException(failure)
                return HTTPException(failure)
            self._logger.trace("Failed to get version: %d %s", failure.status, failure.message)
            if raise_on_error:
                raise HTTPException(failure)
            return HTTPException(failure)

    async def fetch_routeplanner_status(self) -> RoutePlannerStart | HTTPException:
        """|coro|
        Fetches the routeplanner status response from the target node.
        """
        async with self._session.get(
            self.get_endpoint_routeplanner_status(),
            headers={
                "Authorization": self.password,
                "Client-Name": f"PyLav/{self.node_manager.client.lib_version}",
                "App-Id": self.node_manager.client._user_id,
            },
            params={"trace": "true" if self.trace else "false"},
        ) as res:
            if res.status in GOOD_RESPONSE_RANGE:
                data = await res.json(loads=json.loads)
                data["type"] = data["class"]
                del data["class"]
                return from_dict(data_class=RoutePlannerStart, data=data)
            failure = from_dict(data_class=LavalinkError, data=await res.json(loads=json.loads))
            if res.status in [401, 403]:
                raise UnauthorizedException(failure)
            self._logger.trace("Failed to get routeplanner status: %d %s", failure.status, failure.message)
            return HTTPException(failure)

    async def post_routeplanner_free_address(self, address: str) -> None | HTTPException:
        """|coro|
        Frees the given address from the routeplanner.
        """
        async with self._session.post(
            self.get_endpoint_routeplanner_free_address(),
            headers={
                "Authorization": self.password,
                "Client-Name": f"PyLav/{self.node_manager.client.lib_version}",
                "App-Id": self.node_manager.client._user_id,
            },
            json={"address": address},
            params={"trace": "true" if self.trace else "false"},
        ) as res:
            if res.status in GOOD_RESPONSE_RANGE:
                return
            failure = from_dict(data_class=LavalinkError, data=await res.json(loads=json.loads))
            if res.status in [401, 403]:
                raise UnauthorizedException(failure)
            self._logger.trace("Failed to free routeplanner address: %d %s", failure.status, failure.message)
            return HTTPException(failure)

    async def post_routeplanner_free_all(self) -> None | HTTPException:
        """|coro|
        Frees all addresses from the routeplanner.
        """
        async with self._session.post(
            self.get_endpoint_routeplanner_free_all(),
            headers={
                "Authorization": self.password,
                "Client-Name": f"PyLav/{self.node_manager.client.lib_version}",
                "App-Id": self.node_manager.client._user_id,
            },
            params={"trace": "true" if self.trace else "false"},
        ) as res:
            if res.status in GOOD_RESPONSE_RANGE:
                return
            failure = from_dict(data_class=LavalinkError, data=await res.json(loads=json.loads))
            if res.status in [401, 403]:
                raise UnauthorizedException(failure)
            self._logger.trace("Failed to free all routeplanner addresses: %d %s", failure.status, failure.message)
            return HTTPException(failure)

    # REST API - Wrappers

    async def fetch_node_version(self) -> Version:
        """|coro|
        Fetches the version response from the target node.
        """
        self._version = await self.fetch_version(raise_on_error=True)
        return self._version

    async def fetch_api_version(self) -> None:
        """|coro|
        Fetches the API version response from the target node.
        """
        if self.version is None:
            await self.fetch_node_version()

    async def get_guild_player(self, guild_id: int) -> rest_api.LavalinkPlayer:
        """|coro|
        Fetches the player for the given guild ID.
        """
        async with self._session.get(
            self.get_endpoint_session_player_by_guild_id(guild_id=guild_id),
            headers={
                "Authorization": self.password,
                "Client-Name": f"PyLav/{self.node_manager.client.lib_version}",
                "App-Id": self.node_manager.client._user_id,
            },
            params={"trace": "true" if self.trace else "false"},
        ) as res:
            if res.status in GOOD_RESPONSE_RANGE:
                return from_dict(data_class=rest_api.LavalinkPlayer, data=await res.json(loads=json.loads))
            if res.status in [401, 403]:
                raise UnauthorizedException
        raise ValueError(f"Server returned an unexpected return code: {res.status}")

    async def get_track(
        self, query: Query, first: bool = False, bypass_cache: bool = False, sleep: bool = False
    ) -> rest_api.LoadTrackResponses:
        """|coro|
        Gets all tracks associated with the given query.

        Parameters
        ----------
        query: :class:`Query`
            The query to perform a search for.
        first: :class:`bool`
            Whether to return the first result or all results.
        bypass_cache: :class:`bool`
            Whether to bypass the cache.
        sleep: :class:`bool`
            Whether to sleep for 1 second before returning the response.
        Returns
        -------
        LavalinkLoadTrackObjects
            Lavalink LoadTrack Response object
        """
        if not bypass_cache:
            if cached_entry := await self.get_track_from_cache(query=query, first=first):
                return cached_entry
        if (
            self.node_manager.client.local_tracks_cache.is_ready
            and query.is_local
            and f"{query._query}" in self.node_manager.client.local_tracks_cache.path_to_track
        ):
            return self.parse_loadtrack_response(
                {
                    "loadType": "track",
                    "data": self.node_manager.client.local_tracks_cache.path_to_track[f"{query._query}"],
                }
            )
        response = await self.fetch_loadtracks(query=query)
        if sleep:
            await asyncio.sleep(0.05)
        if isinstance(response, HTTPException):
            return response
        if first:
            match response.loadType:
                case "track":
                    return response
                case "search":
                    return rest_api.TrackResponse(loadType="track", data=response.data[0])
                case "playlist":
                    return rest_api.TrackResponse(loadType="track", data=response.data.tracks[0])
        return response

    async def search_youtube_music(self, query: str, bypass_cache: bool = False) -> rest_api.LoadTrackResponses:
        """|coro|
        Gets the query from YouTube music.
        Parameters
        ----------
        query: :class:`str`
            The query to search for.
        bypass_cache: :class:`bool`
            Whether to bypass the cache.
        Returns
        -------
        LavalinkLoadTrackObjects
            Lavalink LoadTrack Response Object
        """
        return await self.get_track(await self._query_cls.from_string(f"ytmsearch:{query}"), bypass_cache=bypass_cache)

    async def search_youtube(self, query: str, bypass_cache: bool = False) -> rest_api.LoadTrackResponses:
        """|coro|
        Gets the query from YouTube music.
        Parameters
        ----------
        query: :class:`str`
            The query to search for.
        bypass_cache: :class:`bool`
            Whether to bypass the cache.
        Returns
        -------
        LavalinkLoadTrackObjects
            Lavalink LoadTrack Response Object
        """
        return await self.get_track(await self._query_cls.from_string(f"ytsearch:{query}"), bypass_cache=bypass_cache)

    async def search_soundcloud(self, query: str, bypass_cache: bool = False) -> rest_api.LoadTrackResponses:
        """|coro|
        Gets the query from Soundcloud.
        Parameters
        ----------
        query: :class:`str`
            The query to search for.
        bypass_cache: :class:`bool`
            Whether to bypass the cache.
        Returns
        -------
        LavalinkLoadTrackObjects
            Lavalink LoadTrack Response Object
        """
        return await self.get_track(await self._query_cls.from_string(f"scsearch:{query}"), bypass_cache=bypass_cache)

    async def search_spotify(self, query: str, bypass_cache: bool = False) -> rest_api.LoadTrackResponses:
        """|coro|
        Gets the query from Spotify.
        Parameters
        ----------
        query: :class:`str`
            The query to search for.
        bypass_cache: :class:`bool`
            Whether to bypass the cache.
        Returns
        -------
        LavalinkLoadTrackObjects
            Lavalink LoadTrack Response Object
        """
        return await self.get_track(await self._query_cls.from_string(f"spsearch:{query}"), bypass_cache=bypass_cache)

    async def search_apple_music(self, query: str, bypass_cache: bool = False) -> rest_api.LoadTrackResponses:
        """|coro|
        Gets the query from Apple Music.
        Parameters
        ----------
        query: :class:`str`
            The query to search for.
        bypass_cache: :class:`bool`
            Whether to bypass the cache.

        Returns
        -------
        LavalinkLoadTrackObjects
            Lavalink LoadTrack Response Object
        """
        return await self.get_track(await self._query_cls.from_string(f"amsearch:{query}"), bypass_cache=bypass_cache)

    async def search_deezer(self, query: str, bypass_cache: bool = False) -> rest_api.LoadTrackResponses:
        """|coro|
        Gets the query from Deezer.
        Parameters
        ----------
        query: :class:`str`
            The query to search for.
        bypass_cache: :class:`bool`
            Whether to bypass the cache.

        Returns
        -------
        LavalinkLoadTrackObjects
            Lavalink LoadTrack Response Object
        """
        return await self.get_track(await self._query_cls.from_string(f"dzsearch:{query}"), bypass_cache=bypass_cache)

    async def search_yandex(self, query: str, bypass_cache: bool = False) -> rest_api.LoadTrackResponses:
        """|coro|
        Gets the query from Yandex Music.
        Parameters
        ----------
        query: :class:`str`
            The query to search for.
        bypass_cache: :class:`bool`
            Whether to bypass the cache.

        Returns
        -------
        LavalinkLoadTrackObjects
            Lavalink LoadTrack Response Object
        """
        return await self.get_track(await self._query_cls.from_string(f"ymsearch:{query}"), bypass_cache=bypass_cache)

    async def get_query_speak(self, query: str, bypass_cache: bool = False) -> rest_api.LoadTrackResponses:
        """|coro|
        Gets the query for speak.
        Parameters
        ----------
        query: :class:`str`
            The query to search for.
        bypass_cache: :class:`bool`
            Whether to bypass the cache.
        Returns
        -------
        LavalinkLoadTrackObjects
            Lavalink LoadTrack Response Object
        """
        return await self.get_track(
            await self._query_cls.from_string(f"speak:{query[:200]}"), bypass_cache=bypass_cache
        )

    async def get_query_localfiles(
        self,
        query: str,
        bypass_cache: bool = True,
    ) -> rest_api.LoadTrackResponses:
        """|coro|
        Gets the query from Localfiles.
        Parameters
        ----------
        query: :class:`str`
            The query to search for.
        bypass_cache: :class:`bool`
            Whether to bypass the cache.

        Returns
        -------
        LavalinkLoadTrackObjects
            Lavalink LoadTrack Response Object
        """
        return await self.get_track(await self._query_cls.from_string(query), bypass_cache=bypass_cache)

    def get_filter_payload(
        self,
        *,
        player: Player,
        volume: Volume = None,
        equalizer: Equalizer = None,
        karaoke: Karaoke = None,
        timescale: Timescale = None,
        tremolo: Tremolo = None,
        vibrato: Vibrato = None,
        rotation: Rotation = None,
        distortion: Distortion = None,
        low_pass: LowPass = None,
        channel_mix: ChannelMix = None,
        echo: Echo = None,
        reset_no_set: bool = False,
        reset: bool = False,
    ) -> JSON_DICT_TYPE:
        """Gets the filter payload."""
        if reset:
            return {}

        payload = {}
        if self.has_filter("volume"):
            self._get_filter_payload_volume(payload, volume)
        if self.has_filter("equalizer"):
            self._get_filter_payload_equalizer(equalizer, payload, player, reset_no_set)
        if self.has_filter("karaoke"):
            self._get_filter_payload_karaoke(karaoke, payload, player, reset_no_set)
        if self.has_filter("timescale"):
            self._get_filter_payload_timescale(payload, player, reset_no_set, timescale)
        if self.has_filter("tremolo"):
            self._get_filter_payload_tremolo(payload, player, reset_no_set, tremolo)
        if self.has_filter("vibrato"):
            self._get_filter_payload_vibrato(payload, player, reset_no_set, vibrato)
        if self.has_filter("rotation"):
            self._get_filter_payload_rotation(payload, player, reset_no_set, rotation)
        if self.has_filter("distortion"):
            self._get_filter_payload_distortion(distortion, payload, player, reset_no_set)
        if self.has_filter("lowPass"):
            self._get_filter_payload_low_pass(low_pass, payload, player, reset_no_set)
        if self.has_filter("channelMix"):
            self._get_filter_payload_channel_mix(channel_mix, payload, player, reset_no_set)
        if self.has_filter("echo"):
            self._get_filter_payload_echo(echo, payload, player, reset_no_set)

        return payload

    @staticmethod
    def _get_filter_payload_volume(payload: JSON_DICT_TYPE, volume: Volume) -> None:
        if volume:
            payload["volume"] = volume.get()

    @staticmethod
    def _get_filter_payload_echo(echo: Echo, payload: JSON_DICT_TYPE, player: Player, reset_no_set: bool) -> None:
        if echo:
            payload["echo"] = echo.get()
        elif not reset_no_set and player.echo:
            payload["echo"] = player.echo.get()

    @staticmethod
    def _get_filter_payload_channel_mix(
        channel_mix: ChannelMix, payload: JSON_DICT_TYPE, player: Player, reset_no_set: bool
    ) -> None:
        if channel_mix:
            payload["channelMix"] = channel_mix.get()
        elif not reset_no_set and player.channel_mix:
            payload["channelMix"] = player.channel_mix.get()

    @staticmethod
    def _get_filter_payload_low_pass(
        low_pass: LowPass, payload: JSON_DICT_TYPE, player: Player, reset_no_set: bool
    ) -> None:
        if low_pass:
            payload["lowPass"] = low_pass.get()
        elif not reset_no_set and player.low_pass:
            payload["lowPass"] = player.low_pass.get()

    @staticmethod
    def _get_filter_payload_distortion(
        distortion: Distortion, payload: JSON_DICT_TYPE, player: Player, reset_no_set: bool
    ) -> None:
        if distortion:
            payload["distortion"] = distortion.get()
        elif not reset_no_set and player.distortion:
            payload["distortion"] = player.distortion.get()

    @staticmethod
    def _get_filter_payload_rotation(
        payload: JSON_DICT_TYPE, player: Player, reset_no_set: bool, rotation: Rotation
    ) -> None:
        if rotation:
            payload["rotation"] = rotation.get()
        elif not reset_no_set and player.rotation:
            payload["rotation"] = player.rotation.get()

    @staticmethod
    def _get_filter_payload_vibrato(
        payload: JSON_DICT_TYPE, player: Player, reset_no_set: bool, vibrato: Vibrato
    ) -> None:
        if vibrato:
            payload["vibrato"] = vibrato.get()
        elif not reset_no_set and player.vibrato:
            payload["vibrato"] = player.vibrato.get()

    @staticmethod
    def _get_filter_payload_tremolo(
        payload: JSON_DICT_TYPE, player: Player, reset_no_set: bool, tremolo: Tremolo
    ) -> None:
        if tremolo:
            payload["tremolo"] = tremolo.get()
        elif not reset_no_set and player.timescale:
            payload["timescale"] = player.timescale.get()

    @staticmethod
    def _get_filter_payload_timescale(
        payload: JSON_DICT_TYPE, player: Player, reset_no_set: bool, timescale: Timescale
    ) -> None:
        if timescale:
            payload["timescale"] = timescale.get()
        elif not reset_no_set and player.timescale:
            payload["timescale"] = player.timescale.get()

    @staticmethod
    def _get_filter_payload_karaoke(
        karaoke: Karaoke, payload: JSON_DICT_TYPE, player: Player, reset_no_set: bool
    ) -> None:
        if karaoke:
            payload["karaoke"] = karaoke.get()
        elif not reset_no_set and player.karaoke:
            payload["karaoke"] = player.karaoke.get()

    @staticmethod
    def _get_filter_payload_equalizer(
        equalizer: Equalizer, payload: JSON_DICT_TYPE, player: Player, reset_no_set: bool
    ) -> None:
        if equalizer:
            payload["equalizer"] = equalizer.get()
        elif not reset_no_set and player.equalizer:
            payload["equalizer"] = player.equalizer.get()

    async def filters(
        self,
        *,
        player: Player,
        volume: Volume = None,
        equalizer: Equalizer = None,
        karaoke: Karaoke = None,
        timescale: Timescale = None,
        tremolo: Tremolo = None,
        vibrato: Vibrato = None,
        rotation: Rotation = None,
        distortion: Distortion = None,
        low_pass: LowPass = None,
        channel_mix: ChannelMix = None,
        echo: Echo = None,
    ) -> None:
        """|coro|
        Set the filters for a player.
        """
        payload = self.get_filter_payload(
            player=self.node_manager.client.player_manager.get(player.guild.id),
            volume=volume,
            equalizer=equalizer,
            karaoke=karaoke,
            timescale=timescale,
            tremolo=tremolo,
            vibrato=vibrato,
            rotation=rotation,
            distortion=distortion,
            low_pass=low_pass,
            channel_mix=channel_mix,
            echo=echo,
        )
        player.add_voice_to_payload(payload)
        await self.patch_session_player(guild_id=player.guild.id, payload={"filters": payload})



from __future__ import annotations

__all__ = (
    "PYLAV_NODES",
    "PYLAV_NODE_SETTINGS",
    "BUNDLED_NODES_IDS_HOST_MAPPING",
    "PYLAV_BUNDLED_NODES_SETTINGS",
)

from collections.abc import Mapping
from typing import Any, NotRequired, TypedDict  # noqa


class NodeYaml(TypedDict):
    """The settings for a node."""

    server: dict[str, Any]
    lavalink: dict[str, Any]


class NodeInfo(TypedDict):
    """The info for a node."""

    unique_identifier: NotRequired[int]
    name: NotRequired[str]
    host: NotRequired[str]
    port: int
    ssl: bool
    password: str
    resume_timeout: int
    reconnect_attempts: int
    search_only: bool
    managed: bool
    disabled_sources: list[str]
    temporary: bool
    extras: dict[str, Any]
    yaml: NodeYaml


# Mapping of the PyLav public nodes
PYLAV_NODES: Mapping[int, tuple[str, tuple[float, float]]] = {
    1: ("london", (1.3213765, 103.6956208)),
    2: ("new_york_city", (40.606872, -74.1769477)),
}

PYLAV_NODE_SETTINGS: NodeInfo = {
    "port": 443,
    "ssl": True,
    "password": "default",
    "resume_timeout": 600,
    "reconnect_attempts": -1,
    "search_only": False,
    "managed": False,
    "disabled_sources": ["local", "http"],
    "temporary": True,
    "extras": {},
    "yaml": {
        "server": {},
        "lavalink": {"server": {"password": "..."}},
    },
}
PYLAV_NODE_SETTINGS["yaml"]["server"]["port"] = PYLAV_NODE_SETTINGS["port"]

PYLAV_LONDON_NODE_SETTINGS: NodeInfo = PYLAV_NODE_SETTINGS.copy()
PYLAV_LONDON_NODE_SETTINGS["host"] = "ll-gb.draper.wtf"
PYLAV_LONDON_NODE_SETTINGS["name"] = "PyLav London (Bundled)"
PYLAV_LONDON_NODE_SETTINGS["unique_identifier"] = 1

PYLAV_NY_NODE_SETTINGS: NodeInfo = PYLAV_NODE_SETTINGS.copy()
PYLAV_NY_NODE_SETTINGS["host"] = "ll-us-ny.draper.wtf"
PYLAV_NY_NODE_SETTINGS["name"] = "PyLav US-NY (Bundled)"
PYLAV_NY_NODE_SETTINGS["unique_identifier"] = 2


BUNDLED_NODES_IDS_HOST_MAPPING: Mapping[int, str] = {
    1: "ll-gb.draper.wtf",
    2: "ll-us-ny.draper.wtf",
    1001: "lava.link",
}

PYLAV_BUNDLED_NODES_SETTINGS: Mapping[str, NodeInfo] = {
    "ll-gb.draper.wtf": PYLAV_LONDON_NODE_SETTINGS,
    "ll-us-ny.draper.wtf": PYLAV_NY_NODE_SETTINGS,
    "lava.link": {
        "host": "lava.link",
        "unique_identifier": 1001,
        "name": "Lava.Link (Bundled)",
        "password": "...",
        "port": 80,
        "resume_timeout": 600,
        "reconnect_attempts": -1,
        "ssl": False,
        "search_only": False,
        "managed": False,
        "disabled_sources": ["local"],
        "temporary": True,
        "extras": {},
        "yaml": {
            "server": {},
            "lavalink": {"server": {"password": "..."}},
        },
    },
}
PYLAV_BUNDLED_NODES_SETTINGS["ll-gb.draper.wtf"]["yaml"]["server"]["address"] = PYLAV_BUNDLED_NODES_SETTINGS[
    "ll-gb.draper.wtf"
]["host"]


PYLAV_BUNDLED_NODES_SETTINGS["ll-us-ny.draper.wtf"]["yaml"]["server"]["address"] = PYLAV_BUNDLED_NODES_SETTINGS[
    "ll-us-ny.draper.wtf"
]["host"]

PYLAV_BUNDLED_NODES_SETTINGS["lava.link"]["yaml"]["server"]["address"] = PYLAV_BUNDLED_NODES_SETTINGS["lava.link"][
    "host"
]
PYLAV_BUNDLED_NODES_SETTINGS["lava.link"]["yaml"]["server"]["port"] = PYLAV_BUNDLED_NODES_SETTINGS["lava.link"]["port"]



from __future__ import annotations

import contextlib
import inspect
import os
import pathlib
import shutil
import sys
from collections.abc import Callable, Iterator
from typing import TypeVar, get_type_hints

from pylav.type_hints.generics import ANY_GENERIC_TYPE, PARAM_SPEC_TYPE

T = TypeVar("T")


def _get_path(path: T | pathlib.Path) -> str | T | None:
    return get_true_path(path, fallback=path)


def update_event_loop_policy() -> None:
    if sys.implementation.name == "cpython":
        # Let's not force this dependency, uvloop is much faster on cpython
        try:
            import uvloop  # type: ignore
        except ImportError:
            pass
        else:
            import asyncio

            if not isinstance(asyncio.get_event_loop_policy(), uvloop.EventLoopPolicy):
                asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())


def check_annotated(
    func: Callable[PARAM_SPEC_TYPE, ANY_GENERIC_TYPE]  # type: ignore
) -> Callable[PARAM_SPEC_TYPE, ANY_GENERIC_TYPE]:  # type: ignore
    hints = get_type_hints(func, include_extras=True)
    spec = inspect.getfullargspec(func)

    def wrapper(*args: PARAM_SPEC_TYPE.args, **kwargs: PARAM_SPEC_TYPE.kwargs) -> ANY_GENERIC_TYPE:
        for idx, arg_name in enumerate(spec[0]):
            hint = hints.get(arg_name)
            validators = getattr(hint, "__metadata__", None)
            if not validators:
                continue
            for validator in validators:
                validator.validate_value(args[idx])

        return func(*args, **kwargs)

    return wrapper


def add_property(inst: object, name: str, method: Callable) -> None:
    cls = type(inst)
    if not hasattr(cls, "__per_instance"):
        cls = type(cls.__name__, (cls,), {})
        cls.__per_instance = True
        inst.__class__ = cls
    setattr(cls, name, property(method))


def get_true_path(executable: str, fallback: ANY_GENERIC_TYPE = None) -> str | ANY_GENERIC_TYPE | None:
    """Returns the true path of the executable."""

    path = os.environ.get("JAVA_HOME", executable)
    with add_env_path(path if os.path.isdir(path) else os.path.split(path)[0]) as path_string:
        executable = shutil.which(executable, path=path_string)
    return executable or fallback


@contextlib.contextmanager
def add_env_path(path: str | os.PathLike) -> Iterator[str]:
    """Adds a path to the environment path temporarily."""
    path = os.fspath(path)
    existing_path = "PATH" in os.environ
    old_path = os.environ["PATH"] if existing_path else None
    try:
        if path not in os.environ["PATH"]:
            yield path + os.pathsep + os.environ["PATH"]
        else:
            yield os.environ["PATH"]
    finally:
        if existing_path:
            os.environ["PATH"] = old_path
        else:
            del os.environ["PATH"]


def fix(string: str, data, *, e: bool = False) -> str | bytes:
    index_list, rem = data
    new_string = string[rem : -1 * rem]
    res = ""
    for idx in range(0, len(new_string)):
        # checking for index list for uppercase
        if idx in index_list:
            res += new_string[idx].upper()
        else:
            res += new_string[idx]
    res += "=="
    final = res
    return final.encode() if e else final



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

from asyncpg import UniqueViolationError  # type: ignore
from piccolo.columns import JSONB, M2M, LazyTableReference, Text
from piccolo.columns.indexes import IndexMethod
from piccolo.table import Table

from pylav.logging import getLogger
from pylav.storage.database.tables.misc import DATABASE_ENGINE

if TYPE_CHECKING:
    from pylav.nodes.api.responses.track import Track

LOGGER = getLogger("PyLav.Database.Track")
_LOCK = asyncio.Lock()


class TrackRow(Table, db=DATABASE_ENGINE, tablename="track"):
    identifier = Text(null=True, default=None, index=True)
    sourceName = Text(null=True, default=None, index=True)
    title = Text(null=True, default=None, index=True, index_method=IndexMethod.gin)
    uri = Text(null=True, default=None, index=True)
    isrc = Text(null=True, default=None, index=True)
    encoded = Text(null=False, index=True, primary_key=True)
    artworkUrl = Text(null=True, default=None)
    info = JSONB(null=True, default=None)
    pluginInfo = JSONB(null=True, default=None)
    queries = M2M(LazyTableReference("TrackToQueries", module_path="pylav.storage.database.tables.m2m"))
    playlists = M2M(LazyTableReference("TrackToPlaylists", module_path="pylav.storage.database.tables.m2m"))

    @classmethod
    async def get_or_create(
        cls,
        track: Track,
    ) -> TrackRow:
        async with _LOCK:
            try:
                kwargs = {
                    TrackRow.identifier: track.info.identifier,
                    TrackRow.sourceName: track.info.sourceName,
                    TrackRow.title: track.info.title,
                    TrackRow.uri: track.info.uri,
                    TrackRow.isrc: track.info.isrc,
                    TrackRow.encoded: track.encoded,
                    TrackRow.artworkUrl: track.info.artworkUrl,
                    TrackRow.info: track.info.to_dict(),
                    TrackRow.pluginInfo: track.pluginInfo.to_dict() if track.pluginInfo else None,
                }

                track = await cls.objects().get_or_create(cls.encoded == track.encoded, kwargs)
                if (not track._was_created) and (
                    (track.info is None and kwargs.get(cls.info._meta.db_column_name))
                    or (track.pluginInfo is None and kwargs.get(cls.pluginInfo._meta.db_column_name))
                ):
                    await cls.update({k._meta.db_column_name: v for k, v in kwargs.items()}).where(
                        cls.encoded == track.encoded
                    )
                return track
            except UniqueViolationError:
                obj = cls(encoded=track.encoded, **{k._meta.db_column_name: v for k, v in kwargs.items()})
                obj._exists_in_db = True
                return obj
            except Exception as e:
                LOGGER.trace("Error while creating track: %s", e, exc_info=True)
                raise e



import dataclasses
from typing import Annotated

from pylav.nodes.api.responses.filters.misc import ValueRange


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class Tremolo:
    frequency: Annotated[float | None, ValueRange(min=0.0, max=float("inf"))] = None
    depth: Annotated[float | None, ValueRange(min=0.0, max=1.0)] = None

    def to_dict(self) -> dict[str, float]:
        response = {}
        if self.frequency is not None:
            response["frequency"] = self.frequency
        if self.depth is not None:
            response["depth"] = self.depth
        return response



from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING, Any

import aiohttp
import aiohttp_client_cache
from yarl import URL

from pylav.compat import json
from pylav.extension.radio.base_url import pick_base_url
from pylav.extension.radio.objects import Codec, Country, CountryCode, Language, State, Station, Tag
from pylav.extension.radio.utils import TransformerCache, type_check
from pylav.logging import getLogger
from pylav.type_hints.dict_typing import JSON_DICT_TYPE
from pylav.utils.vendor.redbot import AsyncIter

if TYPE_CHECKING:
    from pylav.core.client import Client

LOGGER = getLogger("PyLav.extension.RadioBrowser")


class Request:
    """A wrapper for the aiohttp client."""

    __slots__ = ("_headers", "_cached_session", "_session")

    def __init__(
        self,
        headers: dict[str, str] | None = None,
        cached_session: aiohttp_client_cache.CachedSession | None = None,
        session: aiohttp.ClientSession | None = None,
    ) -> None:
        self._headers = headers
        self._cached_session = cached_session
        self._session = session

    async def get(self, url: str | URL, skip_cache: bool = False, **kwargs: Any) -> list[JSON_DICT_TYPE] | None:
        if skip_cache:
            async with self._session.get(url, params=kwargs) as resp:
                if resp.status == 200:
                    return await resp.json(loads=json.loads)
        else:
            async with self._cached_session.get(url, headers=self._headers, params=kwargs) as resp:
                if resp.status == 200:
                    return await resp.json(loads=json.loads)
        resp.raise_for_status()


class RadioBrowser:
    """This class implements the main interface for the Radio Browser API."""

    def __init__(self, client: Client) -> None:
        headers = {"User-Agent": f"PyLav/{client.lib_version}"}
        self._client = client
        self.request = Request(
            headers=headers, cached_session=self._client.cached_session, session=self._client.session
        )
        self._disabled = False

    async def initialize(self) -> None:
        try:
            self._disabled = not await self.base_url
            if self._disabled:
                LOGGER.error("Error while initializing the Radio Browser extension - disabling it")
                return
            LOGGER.debug("Priming radio cache")
            await TransformerCache.fill_cache(self._client)
            await self.stations_by_clicks(limit=25)
            await self.stations_by_votes(limit=25)
            TransformerCache.fill_choice_cache()
            LOGGER.debug("Radio cache primed")
        except Exception as e:
            LOGGER.error("Error while initializing the Radio Browser extension - disabling it")
            LOGGER.debug(e, exc_info=e)
            self._disabled = True

    @property
    async def base_url(self) -> URL:
        """The base URL for the Radio Browser API."""
        if url := await pick_base_url(self._client.session):
            return url
        self._disabled = True
        return URL()

    @property
    def disabled(self) -> bool:
        """Whether the extension is disabled."""
        return self._disabled

    @type_check
    async def countries(self, code: str | None = None) -> list[Country]:
        """Lists all countries.

        Args:
            code (str, optional): Filter by country code. Defaults to None.

        Returns:
            list: Countries.

        See details:
            https://de1.api.radio-browser.info/#List_of_countries
        """
        url = await self.base_url / "json" / "countries"
        if code:
            url /= code
        if self._disabled:
            return []
        return [Country(**country) for country in await self.request.get(url, hidebroken="true")]

    @type_check
    async def countrycodes(self, code: str | None = None) -> list[CountryCode]:
        """Lists all countries.

        Args:
            code (str, optional): Filter by country code. Defaults to None.

        Returns:
            list: Countries.

        See details:
            https://de1.api.radio-browser.info/#List_of_countrycodes
        """
        url = await self.base_url / "json" / "countrycodes"
        if code:
            url /= code
        if self._disabled:
            return []
        return [CountryCode(**country) for country in await self.request.get(url, hidebroken="true")]

    @type_check
    async def codecs(self, codec: str | None = None) -> list[Codec]:
        """Lists all codecs.

        Args:
            codec (str, optional): Filter by codec. Defaults to None.

        Returns:
            list: Codecs.

        See details:
            https://de1.api.radio-browser.info/#List_of_codecs
        """
        url = await self.base_url / "json" / "codecs"
        if self._disabled:
            return []
        response = await self.request.get(url, hidebroken="true")
        if codec:
            return [Codec(**tag) for tag in filter(lambda s: s["name"].lower() == codec.lower(), response)]
        return [Codec(**tag) for tag in response]

    @type_check
    async def states(self, country: str | None = None, state: str | None = None) -> list[State]:
        """Lists all states.

        Args:
            country (str, optional): Filter by country. Defaults to None.
            state (str, optionla): Filter by state. Defaults to None.

        Returns:
            list: States.

        See details:
            https://de1.api.radio-browser.info/#List_of_states
        """
        url = await self.base_url / "json" / "states"
        if self._disabled:
            return []
        response = await self.request.get(url, hidebroken="true")

        if country:
            if state:
                return [
                    State(**state)
                    for state in filter(
                        lambda s: s["country"].lower() == country.lower() and s["name"].lower() == state.lower(),
                        response,
                    )
                ]
            return [State(**state) for state in filter(lambda s: s["country"].lower() == country.lower(), response)]
        if state:
            return [State(**state) for state in filter(lambda s: s["name"].lower() == state.lower(), response)]
        return [State(**state) for state in response]

    @type_check
    async def languages(self, language: str | None = None) -> list[Language]:
        """Lists all languages.

        Args:
            language (str, optional): Filter by language. Defaults to None.

        Returns:
            list: Languages.

        See details:
            https://de1.api.radio-browser.info/#List_of_languages
        """
        url = await self.base_url / "json" / "languages"
        if language:
            url /= language
        if self._disabled:
            return []
        response = await self.request.get(url, hidebroken="true")
        return [Language(**language) for language in response]

    @type_check
    async def tags(self, tag: str | None = None) -> list[Tag]:
        """Lists all tags.

        Args:
            tag (str, optional): Filter by tag. Defaults to None.

        Returns:
            list: Tags.

        See details:
            https://de1.api.radio-browser.info/#List_of_tags
        """
        url = await self.base_url / "json" / "tags"
        if tag:
            url /= tag.lower()
        if self._disabled:
            return []
        response = await self.request.get(url, hidebroken="true")
        return [Tag(**tag) for tag in response]

    async def station_by_uuid(self, stationuuid: str) -> list[Station]:
        """Radio station by stationuuid.

        Args:
            stationuuid (str): A globally unique identifier for the station.

        Returns:
            list: Stations.

        See details:
            https://de1.api.radio-browser.info/#List_of_radio_stations
        """
        url = await self.base_url / "json" / "stations" / "byuuid" / stationuuid
        if self._disabled:
            return []
        response = await self.request.get(url, hidebroken="true")
        return [Station(radio_api_client=self, **station) async for station in AsyncIter(response, steps=250)]

    async def stations_by_name(
        self, name: str, exact: bool = False, **kwargs: str | int | bool | None
    ) -> list[Station]:
        """Lists all radio stations by name.

        Args:
            name (str): The name of the station.
            exact (bool): if the search should search for a station with the exact `name`

        Returns:
            list: Stations.

        See details:
            https://de1.api.radio-browser.info/#List_of_radio_stations
        """
        kwargs |= {"name": name, "name_exact": exact, "order": "votes"}
        kwargs["hidebroken"] = kwargs.pop("hidebroken", True)
        return await self.search(**kwargs)

    async def stations_by_codec(
        self, codec: str, exact: bool = False, **kwargs: str | int | bool | None
    ) -> list[Station]:
        """Lists all radio stations by codec.

        Args:
            codec (str): The name of the codec.
            exact (bool): if the search should search for a station with the exact `codec`

        Returns:
            list: Stations.

        See details:
            https://de1.api.radio-browser.info/#List_of_radio_stations
        """
        kwargs |= {"codec": codec, "codec_exact": exact, "order": "votes"}
        kwargs["hidebroken"] = kwargs.pop("hidebroken", True)
        return await self.search(**kwargs)

    async def stations_by_country(
        self, country: str, exact: bool = False, **kwargs: str | int | bool | None
    ) -> list[Station]:
        """Lists all radio stations by country.

        Args:
            country (str): The name of the country.
            exact (bool): if the search should search for a station with the exact `country`

        Returns:
            list: Stations.

        See details:
            https://de1.api.radio-browser.info/#List_of_radio_stations
        """
        kwargs |= {"country": country, "country_exact": exact, "order": "votes"}
        kwargs["hidebroken"] = kwargs.pop("hidebroken", True)
        return await self.search(**kwargs)

    async def stations_by_countrycode(self, code: str, **kwargs: str | int | bool | None) -> list[Station]:
        """Lists all radio stations by country code.

        Arguments:
        ----------
            code (str): Official countrycodes as in ISO 3166-1 alpha-2.
            exact (bool): if the search should search for a station with the exact `code`

        Returns:
            list: Stations.

        See details:
            https://de1.api.radio-browser.info/#List_of_radio_stations
        """
        kwargs |= {"countrycode": code, "order": "votes"}
        kwargs["hidebroken"] = kwargs.pop("hidebroken", True)
        return await self.search(**kwargs)

    async def stations_by_state(
        self, state: str, exact: bool = False, **kwargs: str | int | bool | None
    ) -> list[Station]:
        """Lists all radio stations by state.

        Args:
            state (str): The name of the state.
            exact (bool): if the search should search for a station with the exact `state`

        Returns:
            list: Stations.

        See details:
            https://de1.api.radio-browser.info/#List_of_radio_stations
        """
        kwargs |= {"state": state, "state_exact": exact, "order": "votes"}
        kwargs["hidebroken"] = kwargs.pop("hidebroken", True)
        return await self.search(**kwargs)

    async def stations_by_language(
        self, language: str, exact: bool = False, **kwargs: str | int | bool | None
    ) -> list[Station]:
        """Lists all radio stations by language.

        Args:
            language (str): The name of the language.
            exact (bool): if the search should search for a station with the exact `language`

        Returns:
            list: Stations.

        See details:
            https://de1.api.radio-browser.info/#List_of_radio_stations
        """
        kwargs |= {"language": language, "language_exact": exact, "order": "votes"}
        kwargs["hidebroken"] = kwargs.pop("hidebroken", True)
        return await self.search(**kwargs)

    async def stations_by_tag(self, tag: str, exact: bool = False, **kwargs: str | int | bool | None) -> list[Station]:
        """Lists all radio stations by tag.

        Args:
            tag (str): The name of the tag.
            exact (bool): If set to True, the tag must be exactly the same as the given tag.
        Returns:
            list: Stations.
        See details:
            https://de1.api.radio-browser.info/#List_of_radio_stations
        """
        kwargs |= {"tag": tag, "tag_exact": exact, "order": "votes"}
        kwargs["hidebroken"] = kwargs.pop("hidebroken", True)
        return await self.search(**kwargs)

    async def stations_by_tag_list(self, tag_list: list[str], **kwargs: str | int | bool | None) -> list[Station]:
        """Lists all radio stations by tag. Must match all tags exactly.

        Args:
            tag_list (list): A list of names of tags.

        Returns:
            list: Stations.
        See details:
            https://de1.api.radio-browser.info/#List_of_radio_stations
        """
        tag_list = ",".join(tag_list)
        kwargs["tag_list"] = tag_list
        kwargs |= {"tag_list": tag_list, "order": "votes"}
        kwargs["hidebroken"] = kwargs.pop("hidebroken", True)
        return await self.search(**kwargs)

    async def click_counter(self, stationuuid: str) -> list[dict[str, str | int | bool]]:
        """Increase the click count of a station by one.

        This should be called everytime when a user starts
        playing a stream to mark the stream more popular than others.
        Every call to this endpoint from the same IP address and for
        the same station only gets counted once per day. The call will
        return detailed information about the stat stream, supported output
        formats: JSON

        Args:
            stationuuid (str): A globally unique identifier for the station.

        Returns:
            dict: A dict containing informations about the radio station.

        See details:
            https://de1.api.radio-browser.info/#Count_station_click
        """
        url = await self.base_url / "json" / "url" / f"{stationuuid}"
        return [] if self._disabled else await self.request.get(url, hidebroken="true")

    async def stations(self, **kwargs: str | int | bool | None) -> list[Station]:
        """Lists all radio stations.

        Returns:
            list: Stations.

        See details:
            https://nl1.api.radio-browser.info/#List_of_all_radio_stations
        """
        url = await self.base_url / "json" / "stations"
        if self._disabled:
            return []
        kwargs["hidebroken"] = kwargs.pop("hidebroken", "true")
        return [
            Station(radio_api_client=self, **station)
            async for station in AsyncIter(await self.request.get(url, **kwargs), steps=250)
        ]

    async def stations_by_votes(self, limit: int, **kwargs: str | int | bool | None) -> list[Station]:
        """A list of the highest-voted stations.

        Args:
            limit: Number of wanted stations

        Returns:
            list: Stations.

        See details:
            https://nl1.api.radio-browser.info/#Stations_by_votes
        """
        url = await self.base_url / "json" / "stations" / "topvote" / f"{limit}"
        if self._disabled:
            return []
        kwargs["hidebroken"] = kwargs.pop("hidebroken", "true")
        response = await self.request.get(url, **kwargs)
        return [Station(radio_api_client=self, **station) async for station in AsyncIter(response, steps=250)]

    async def stations_by_clicks(self, limit: int, **kwargs: str | int | bool | None) -> list[Station]:
        """A list of the stations that are clicked the most.

        Args:
            limit: Number of wanted stations

        Returns:
            list: Stations.

        See details:
            https://nl1.api.radio-browser.info/#Stations_by_clicks
        """
        url = await self.base_url / "json" / "stations" / "topclick" / f"{limit}"
        if self._disabled:
            return []
        kwargs["hidebroken"] = kwargs.pop("hidebroken", "true")
        response = await self.request.get(url, **kwargs)
        return [Station(radio_api_client=self, **station) async for station in AsyncIter(response, steps=250)]

    @type_check
    async def search(self, **kwargs: str | int | bool | None) -> list[Station]:
        """Advanced search.

        It will search for the station whose attribute
        contains the search term.

        Arguments:
        ----------
            name (str, optional): Name of the station.
            name_exact (bool, optional): Only exact matches, otherwise all
                matches (default: False).
            country (str, optional): Country of the station.
            country_exact (bool, optional): Only exact matches, otherwise
                all matches (default: False).
            countrycode (str, optional): 2-digit countrycode of the station
                (see ISO 3166-1 alpha-2)
            state (str, optional): State of the station.
            state_exact (bool, optional): Only exact matches, otherwise all
                matches. (default: False)
            language (str, optional): Language of the station.
            language_exact (bool, optional): Only exact matches, otherwise
                all matches. (default: False)
            tag (str, optional): Tag of the station.
            tag_exact (bool, optional): Only exact matches, otherwise all
                matches. (default: False)
            tag_list (str, optional): A comma-separated list of tag.
            bitrate_min (int, optional): Minimum of kbps for bitrate field of
                stations in result. (default: 0)
            bitrate_max (int, optional): Maximum of kbps for bitrate field of
                stations in result. (default: 1000000)
            order (str, optional): The result list will be sorted by: name,
                url, homepage, favicon, tags, country, state, language, votes,
                codec, bitrate, lastcheckok, lastchecktime, clicktimestamp,
                clickcount, clicktrend, random
            reverse (bool, optional): Reverse the result list if set to true.
                (default: false)
            offset (int, optional): Starting value of the result list from
                the database. For example, if you want to do paging on the
                server side. (default: 0)
            limit (int, optional): Number of returned datarows (stations)
                starting with offset (default 100000)
            hidebroken (bool, optional): do list/not list broken stations.
                Note: Not documented in the "Advanced Station Search".

        Returns:
            list: Station.

        See details:
            https://de1.api.radio-browser.info/#Advanced_station_search
        """
        url = await self.base_url / "json" / "stations" / "search"
        if self._disabled:
            return []
        # lowercase tag reference since the API turned to be case-sensitive
        for paramkey in ["tag", "tagList"]:
            if paramkey in kwargs:
                kwargs[paramkey] = kwargs[paramkey].lower()
        kwargs["hidebroken"] = kwargs.pop("hidebroken", "true")
        if kwargs["hidebroken"] is False:
            kwargs["hidebroken"] = "false"
        return [
            Station(radio_api_client=self, **station)
            async for station in AsyncIter(await self.request.get(url, **kwargs), steps=250)
        ]

    async def click(self, station: Station | None = None, station_id: str | None = None) -> None:
        """Increase the click count of a station by one.

        This should be called everytime when a user starts playing a stream to mark the stream more popular than others.
        Every call to this endpoint from the same IP address and for the same station only gets counted once per day.

        Parameters:
            station (Station, optional): The station to click.
            station_id (str, optional): The station uuid to click.
        """
        if (not station) and (not station_id):
            return
        if station:
            station_id = station.stationuuid
        url = await self.base_url / "json" / "url" / f"{station_id}"
        if self._disabled:
            return
        with contextlib.suppress(Exception):
            await self.request.get(url)



from __future__ import annotations

from asyncpg import Connection

from pylav.storage.migrations.logging import LOGGER


async def low_level_v_1_7_0_migration(con: Connection) -> None:
    """Run the low level migration for PyLav 1.7.0."""
    await low_level_v_1_7_0_tracks(con)
    await low_level_v_1_7_0_queries(con)


async def low_level_v_1_7_0_tracks(con: Connection) -> None:
    """Run the tracks migration for PyLav 1.7.0."""
    await run_tracks_migration_v_1_7_0(con)


async def low_level_v_1_7_0_queries(con: Connection) -> None:
    """Run the queries migration for PyLav 1.7.0."""
    await run_query_migration_v_1_7_0(con)


async def run_tracks_migration_v_1_7_0(con: Connection) -> None:
    """
    Add the info and pluginInfo columns to the tracks table.
    """
    has_column = """
        SELECT EXISTS (SELECT 1
        FROM information_schema.columns
        WHERE table_name='version' AND column_name='version')
        """
    has_version_column = await con.fetchval(has_column)
    if not has_version_column:
        return

    version = await con.fetchval("SELECT version from version;")
    if version is None:
        return

    has_column = """
            SELECT EXISTS (SELECT 1
            FROM information_schema.columns
            WHERE table_name='track' AND column_name='info')
            """
    has_column_response = await con.fetchval(has_column)
    if not has_column_response:
        LOGGER.info("----------- Migrating Tracks to PyLav 1.7.0 ---------")
        alter_table = """
        ALTER TABLE IF EXISTS track
        ADD COLUMN IF NOT EXISTS "info" jsonb,
        ADD COLUMN IF NOT EXISTS "pluginInfo" jsonb
        """
        await con.execute(alter_table)


async def run_query_migration_v_1_7_0(con: Connection) -> None:
    """
    Add the pluginInfo column to the query table.
    """
    has_column = """
        SELECT EXISTS (SELECT 1
        FROM information_schema.columns
        WHERE table_name='version' AND column_name='version')
        """
    has_version_column = await con.fetchval(has_column)
    if not has_version_column:
        return

    version = await con.fetchval("SELECT version from version;")
    if version is None:
        return

    has_column = """
            SELECT EXISTS (SELECT 1
            FROM information_schema.columns
            WHERE table_name='query' AND column_name='pluginInfo')
            """
    has_column_response = await con.fetchval(has_column)
    if not has_column_response:
        LOGGER.info("----------- Migrating Query to PyLav 1.7.0 ---------")
        alter_table = """
        ALTER TABLE IF EXISTS query
        ADD COLUMN IF NOT EXISTS "pluginInfo" jsonb
        """
        await con.execute(alter_table)



from __future__ import annotations

from typing import TYPE_CHECKING

from pylav.constants.node import NODE_DEFAULT_SETTINGS
from pylav.storage.migrations.logging import LOGGER

if TYPE_CHECKING:
    from pylav.core.client import Client


async def update_managed_node_settings(client: Client) -> None:
    """Update the managed node settings."""
    LOGGER.info("Running migration - Updating Managed Node Settings")
    # noinspection PyProtectedMember
    config = client._node_config_manager.bundled_node_config()
    data = await config.fetch_yaml()
    data["server"]["undertow"] = NODE_DEFAULT_SETTINGS["server"]["undertow"]
    data["server"]["compression"] = NODE_DEFAULT_SETTINGS["server"]["compression"]
    await config.update_yaml(data)



from __future__ import annotations

from typing import TYPE_CHECKING

from pylav import VERSION
from pylav.storage.migrations.logging import LOGGER

if TYPE_CHECKING:
    from pylav.core.client import Client


async def set_current_version(client: Client) -> None:
    """Set the current version in the database."""
    LOGGER.info("Running migration cleanup")
    await client.lib_db_manager.update_bot_dv_version(VERSION)



from __future__ import annotations

import multiprocessing
import secrets

from pylav import __version__

MAX_SUPPORTED_API_MAJOR_VERSION = 4
TRACK_VERSION = 3
__CPU_COUNT = multiprocessing.cpu_count()

# noinspection SpellCheckingInspection
NODE_DEFAULT_SETTINGS = {
    "server": {
        "port": 2154,
        "address": "localhost",
        "undertow": {
            "threads": {"io": __CPU_COUNT * 2, "worker": __CPU_COUNT * 64},
        },
        "compression": {"enabled": False},
    },
    "spring": {
        "mvc": {
            "async": {
                "request-timeout": -1,
            }
        }
    },
    "lavalink": {
        "plugins": [
            {
                "dependency": "com.github.TopiSenpai.LavaSrc:lavasrc-plugin:3.2.2",
                "repository": "https://maven.topi.wtf/releases",
            },
            {
                "dependency": "com.dunctebot:skybot-lavalink-plugin:1.4.1",
                "repository": "https://m2.duncte123.dev/releases",
            },
            {"dependency": "com.github.topisenpai:sponsorblock-plugin:2.0.0", "repository": "https://jitpack.io"},
            {"dependency": "me.rohank05:lavalink-filter-plugin:0.0.1", "repository": "https://jitpack.io"},
            {"dependency": "com.github.esmBot:lava-xm-plugin:v0.2.1", "repository": "https://jitpack.io"},
        ],
        "server": {
            "password": secrets.token_urlsafe(32),
            "sources": {
                "youtube": True,
                "bandcamp": True,
                "soundcloud": True,
                "twitch": True,
                "vimeo": True,
                "http": True,
                "local": True,
            },
            "filters": {
                "volume": True,
                "equalizer": True,
                "karaoke": True,
                "timescale": True,
                "tremolo": True,
                "vibrato": True,
                "distortion": True,
                "rotation": True,
                "channelMix": True,
                "lowPass": True,
                "echo": True,
            },
            "bufferDurationMs": 400,
            "frameBufferDurationMs": 1000,
            "trackStuckThresholdMs": 10000,
            "youtubePlaylistLoadLimit": 100,
            "opusEncodingQuality": 10,
            "resamplingQuality": "HIGH",
            "useSeekGhosting": True,
            "playerUpdateInterval": 30,
            "youtubeSearchEnabled": True,
            "soundcloudSearchEnabled": True,
            "gc-warnings": True,
            "ratelimit": {
                "ipBlocks": [],
                "excludedIps": [],
                "strategy": "RotateOnBan",
                "searchTriggersFail": True,
                "retryLimit": -1,
            },
            "youtubeConfig": {
                "email": "",
                "password": "",
            },
            "httpConfig": {"proxyHost": "", "proxyPort": 0, "proxyUser": "", "proxyPassword": ""},
        },
    },
    "plugins": {
        "lavasrc": {
            "providers": [
                "dzisrc:%ISRC%",
                'ytmsearch:"%ISRC%"',
                'ytsearch:"%ISRC%"',
                "dzsearch:%QUERY%",
                "ytmsearch:%QUERY%",
                "ytsearch:%QUERY%",
                "scsearch:%QUERY%",
            ],
            "sources": {"spotify": False, "applemusic": True, "deezer": False, "yandexmusic": False},
            "spotify": {
                "clientId": "",
                "clientSecret": "",
                "countryCode": "US",
                "playlistLoadLimit": 110,
                "albumLoadLimit": 220,
            },
            "applemusic": {"countryCode": "US", "mediaAPIToken": None, "playlistLoadLimit": 30, "albumLoadLimit": 30},
            "deezer": {"masterDecryptionKey": ""},
            "yandexmusic": {"accessToken": ""},
        },
        "dunctebot": {
            "ttsLanguage": "en-US",
            "sources": {
                "getyarn": True,
                "clypit": True,
                "tts": True,
                "pornhub": True,
                "reddit": True,
                "ocremix": True,
                "tiktok": True,
                "mixcloud": True,
                "soundgasm": True,
            },
        },
    },
    "metrics": {"prometheus": {"enabled": False, "endpoint": "/metrics"}},
    "sentry": {"dsn": "", "environment": "", "tags": {"pylav_version": __version__}},
    "logging": {
        "file": {"path": "./logs/"},
        "logback": {
            "rollingpolicy": {"max-history": 7, "max-file-size": "25MB", "total-size-cap": "1GB"},
        },
        "level": {"root": "INFO", "lavalink": "INFO"},
        "request": {
            "enabled": True,
            "includeClientInfo": True,
            "includeHeaders": False,
            "includeQueryString": True,
            "includePayload": True,
            "maxPayloadLength": 10000,
        },
    },
}
GOOD_RESPONSE_RANGE = range(200, 299)
JAR_SERVER_RELEASES = "https://api.github.com/repos/lavalink-devs/Lavalink/releases"



from __future__ import annotations

import pathlib

try:
    from redbot.core.i18n import Translator  # type: ignore

    _ = Translator("PyLav", pathlib.Path(__file__))
except ImportError:
    Translator = None

    def _(string: str) -> str:
        return string


def shorten_string(string: str, max_length: int, right: bool = True) -> str:
    """Shortens the given string to the given max length and adds an ellipsis."""
    if not string:
        return string
    if len(string) > max_length:
        if right:
            return string[: max_length - 1] + "\N{HORIZONTAL ELLIPSIS}"
        else:
            return string[(max_length - 1) * -1 :] + "\N{HORIZONTAL ELLIPSIS}"
    return string


def format_time_dd_hh_mm_ss(duration: int | float) -> str:
    """Formats to the given time in milliseconds into DD:HH:MM:SS"""
    seconds = int(duration // 1000)
    if seconds == 0:
        return _("Calculating...")
    days, seconds = divmod(seconds, 24 * 60 * 60)
    hours, seconds = divmod(seconds, 60 * 60)
    minutes, seconds = divmod(seconds, 60)
    day = f"{days:02d}:" if days else ""
    hour = f"{hours:02d}:" if hours or day else ""
    minutes = f"{minutes:02d}:"
    sec = f"{seconds:02d}"
    return f"{day}{hour}{minutes}{sec}"


def format_time_string(seconds: int | float) -> str:
    """Formats the given seconds into a time string

    Examples:
        60 -> 1 minute
        3600 -> 1 hour
        3601 -> 1 hour 1 second
        86400 -> 1 day
        86401 -> 1 day 1 second
    """
    seconds = int(seconds)
    m, s = divmod(seconds, 60)
    h, m = divmod(m, 60)
    d, h = divmod(h, 24)

    hour = _("hour")
    hours = _("hours")

    minute = _("minute")
    minutes = _("minutes")

    second = _("second")
    seconds = _("seconds")

    day = _("day")
    days = _("days")

    if d > 0:
        return f"{d} {day if d == 1 else days} {h} {hour if h == 1 else hours}"

    elif d == 0 and h > 0:
        return f"{h} {hour if h == 1 else hours} {m} {minute if m == 1 else minutes}"

    elif d == 0 and h == 0 and m > 0:
        return f"{m} {minute if m == 1 else minutes} {s} {second if s == 1 else seconds}"

    elif d == 0 and h == 0 and m == 0 and s >= 0:
        return f"{s} {second if s == 1 else seconds}"
    else:
        return ""



from __future__ import annotations

from pathlib import Path

import discord
from redbot.core.i18n import Translator

from pylav.constants.node_features import SUPPORTED_FEATURES, SUPPORTED_SOURCES
from pylav.helpers.format.strings import shorten_string
from pylav.storage.models.node.real import Node

_ = Translator("PyLav", Path(__file__))


class SourceOption(discord.SelectOption):
    def __init__(self, name: str, description: str | None, value: str):
        super().__init__(
            label=shorten_string(max_length=100, string=name),
            description=shorten_string(max_length=100, string=description),
            value=value,
        )


SOURCE_OPTIONS = [
    SourceOption(name=source, description=None, value=source) for source in SUPPORTED_SOURCES.union(SUPPORTED_FEATURES)
]


class NodeOption(discord.SelectOption):
    @classmethod
    async def from_node(cls, node: Node, index: int):
        data = await node.fetch_all()
        return cls(
            label=shorten_string(max_length=100, string=f"{index + 1}. {data['name']}"),
            description=shorten_string(
                max_length=100,
                string=_(
                    "Identifier: {node_identifier_variable_do_not_translate} || SSL: {node_ssl_variable_do_not_translate} || Search-only: {node_config_variable_do_not_translate}"
                ).format(
                    node_identifier_variable_do_not_translate=node.id,
                    node_ssl_variable_do_not_translate=data["ssl"],
                    node_config_variable_do_not_translate=data["search_only"],
                ),
            ),
            value=f"{node.id}",
        )



from __future__ import annotations

import contextlib
from collections import defaultdict
from typing import TYPE_CHECKING

import asyncpg
from asyncpg import Connection
from dacite import from_dict
from packaging.version import parse

from pylav.compat import json
from pylav.constants.playlists import BUNDLED_PLAYLIST_IDS
from pylav.constants.versions import VERSION_1_0_0
from pylav.nodes.api.responses.track import Track
from pylav.players.tracks.decoder import decode_track
from pylav.storage.database.tables.config import LibConfigRow
from pylav.storage.database.tables.nodes import NodeRow
from pylav.storage.database.tables.players import PlayerRow
from pylav.storage.database.tables.playlists import PlaylistRow
from pylav.storage.database.tables.queries import QueryRow
from pylav.storage.database.tables.tracks import TrackRow

if TYPE_CHECKING:
    from pylav.storage.controllers.config import ConfigController


async def run_playlist_migration_v_1_0_0(connection: Connection) -> list[asyncpg.Record]:
    """
    Runs playlist migration 100.
    """
    HAS_COLUMN = """
    SELECT EXISTS (SELECT 1
    FROM information_schema.columns
    WHERE table_name='playlist' AND column_name='tracks')
    """
    has_playlist_tracks = await connection.fetchval(HAS_COLUMN)
    if has_playlist_tracks:
        data = await connection.fetch("SELECT * FROM playlist;")
        await connection.execute("DROP TABLE playlist;")
        return data


async def run_query_migration_v_1_0_0(connection: Connection) -> list[asyncpg.Record] | None:
    """
    Runs playlist migration 100.
    """
    HAS_COLUMN = """
    SELECT EXISTS (SELECT 1
    FROM information_schema.columns
    WHERE table_name='query' AND column_name='tracks')
    """
    has_query_tracks = await connection.fetchval(HAS_COLUMN)
    if has_query_tracks:
        data = await connection.fetch("SELECT * FROM query;")
        await connection.execute("DROP TABLE query;")
        return data


async def run_player_config_v_1_0_0(connection: Connection) -> list[asyncpg.Record]:
    """
    Migrates player config.
    """
    has_column = """
    SELECT EXISTS (SELECT 1
    FROM information_schema.columns
    WHERE table_name='version' AND column_name='version')
    """
    has_version_column = await connection.fetchval(has_column)
    if not has_version_column:
        return []
    version = await connection.fetchval("SELECT version from version;")
    if version is None:
        return []

    version = parse(version)
    if (not version) or version < VERSION_1_0_0:
        return await connection.fetch("SELECT * FROM player;")
    return []


async def run_node_config_v_1_0_0(connection: Connection) -> list[asyncpg.Record]:
    """
    Migrates player config.
    """
    has_column = """
    SELECT EXISTS (SELECT 1
    FROM information_schema.columns
    WHERE table_name='version' AND column_name='version')
    """
    has_version_column = await connection.fetchval(has_column)
    if not has_version_column:
        return []
    version = await connection.fetchval("SELECT version from version;")
    if version is None:
        return []

    version = parse(version)
    if (not version) or version < VERSION_1_0_0:
        return await connection.fetch("SELECT * FROM node;")
    return []


async def run_lib_config_v_1_0_0(connection: Connection) -> list[asyncpg.Record]:
    """
    Migrates player config.
    """
    has_column = """
    SELECT EXISTS (SELECT 1
    FROM information_schema.columns
    WHERE table_name='version' AND column_name='version')
    """
    has_version_column = await connection.fetchval(has_column)
    if not has_version_column:
        return []
    version = await connection.fetchval("SELECT version from version;")
    if version is None:
        return []

    version = parse(version)
    if (not version) or version < VERSION_1_0_0:
        return await connection.fetch("SELECT * FROM lib_config;")


async def migrate_playlists_v_1_0_0(playlists: list[asyncpg.Record]) -> None:
    """Runs playlist migration for version 1.0.0."""
    for playlist in playlists:
        if playlist["id"] in BUNDLED_PLAYLIST_IDS:
            continue
        defaults = {
            PlaylistRow.name: playlist["name"],
            PlaylistRow.scope: playlist["scope"],
            PlaylistRow.author: playlist["author"],
            PlaylistRow.url: playlist["url"],
        }
        playlist_row = await PlaylistRow.objects().get_or_create(PlaylistRow.id == playlist["id"], defaults)
        # noinspection PyProtectedMember
        if not playlist_row._was_created:
            await PlaylistRow.update(defaults).where(PlaylistRow.id == playlist["id"])
        new_tracks = []
        # TODO: Optimize this, after https://github.com/piccolo-orm/piccolo/discussions/683 is answered or fixed
        tracks = json.loads(playlist["tracks"]) if playlist["tracks"] else []
        _temp = defaultdict(list)
        for x in tracks:
            _temp[type(x)].append(x)
        for entry_type, entry_list in _temp.items():
            if entry_type == str:
                for track in entry_list:
                    with contextlib.suppress(Exception):
                        # TODO: Make an API call to the public node?
                        new_tracks.append(await TrackRow.get_or_create(decode_track(track)))
            elif entry_type == dict:
                for track_object in entry_list:
                    new_tracks.append(await TrackRow.get_or_create(from_dict(data_class=Track, data=track_object)))
            else:
                for track_object in entry_list:
                    new_tracks.append(await TrackRow.get_or_create(track_object))
        if new_tracks:
            await playlist_row.add_m2m(*new_tracks, m2m=PlaylistRow.tracks)


async def migrate_queries_v_1_0_0(queries: list[asyncpg.Record]) -> None:
    """Processes queries for migration to version 1.0.0."""
    for query in queries:
        defaults = {QueryRow.name: query["name"]}
        query_row = await QueryRow.objects().get_or_create(QueryRow.identifier == query["identifier"], defaults)

        # noinspection PyProtectedMember
        if not query_row._was_created:
            await QueryRow.update(defaults).where(QueryRow.identifier == query["identifier"])
        new_tracks = []
        # TODO: Optimize this, after https://github.com/piccolo-orm/piccolo/discussions/683 is answered or fixed
        tracks = json.loads(query["tracks"]) if query["tracks"] else []
        _temp = defaultdict(list)
        for x in tracks:
            _temp[type(x)].append(x)
        for entry_type, entry_list in _temp.items():
            if entry_type == str:
                for track in entry_list:
                    with contextlib.suppress(Exception):
                        # TODO: Make an API call to the public node?
                        new_tracks.append(await TrackRow.get_or_create(decode_track(track)))
            elif entry_type == dict:
                for track_object in entry_list:
                    new_tracks.append(await TrackRow.get_or_create(from_dict(data_class=Track, data=track_object)))
            else:
                for track_object in entry_list:
                    new_tracks.append(await TrackRow.get_or_create(track_object))
        if new_tracks:
            await query_row.add_m2m(*new_tracks, m2m=QueryRow.tracks)


async def migrate_player_config_v_1_0_0(players: list[asyncpg.Record]) -> None:
    """Processes player config for migration to version 1.0.0."""
    bulk_insert = []
    for player in players:
        data = {
            "id": player["id"],
            "bot": player["bot"],
            "volume": player["volume"],
            "max_volume": player["max_volume"],
            "auto_play_playlist_id": player["auto_play_playlist_id"],
            "text_channel_id": player["text_channel_id"],
            "notify_channel_id": player["notify_channel_id"],
            "forced_channel_id": player["forced_channel_id"],
            "repeat_current": player["repeat_current"],
            "repeat_queue": player["repeat_queue"],
            "shuffle": player["shuffle"],
            "auto_shuffle": player["auto_shuffle"],
            "auto_play": player["auto_play"],
            "self_deaf": player["self_deaf"],
            "empty_queue_dc": json.loads(player["empty_queue_dc"]),
            "alone_dc": json.loads(player["alone_dc"]),
            "alone_pause": json.loads(player["alone_pause"]),
            "extras": json.loads(player["extras"]),
            "effects": json.loads(player["effects"]),
            "dj_users": player["dj_users"],
            "dj_roles": player["dj_roles"],
        }

        if player["id"] == 0:
            data = {
                PlayerRow.volume: player["volume"],
                PlayerRow.max_volume: player["max_volume"],
                PlayerRow.auto_play_playlist_id: player["auto_play_playlist_id"],
                PlayerRow.text_channel_id: player["text_channel_id"],
                PlayerRow.notify_channel_id: player["notify_channel_id"],
                PlayerRow.forced_channel_id: player["forced_channel_id"],
                PlayerRow.repeat_current: player["repeat_current"],
                PlayerRow.repeat_queue: player["repeat_queue"],
                PlayerRow.shuffle: player["shuffle"],
                PlayerRow.auto_shuffle: player["auto_shuffle"],
                PlayerRow.auto_play: player["auto_play"],
                PlayerRow.self_deaf: player["self_deaf"],
                PlayerRow.empty_queue_dc: json.loads(player["empty_queue_dc"]),
                PlayerRow.alone_dc: json.loads(player["alone_dc"]),
                PlayerRow.alone_pause: json.loads(player["alone_pause"]),
                PlayerRow.extras: json.loads(player["extras"]),
                PlayerRow.effects: json.loads(player["effects"]),
                PlayerRow.dj_users: player["dj_users"],
                PlayerRow.dj_roles: player["dj_roles"],
            }

            playerobj = await PlayerRow.objects().get_or_create(
                (PlayerRow.id == player["id"]) & (PlayerRow.bot == player["bot"]), defaults=data
            )
            if not playerobj._was_created:
                await PlayerRow.update(data).where((PlayerRow.id == player["id"]) & (PlayerRow.bot == player["bot"]))

        else:
            bulk_insert.append(PlayerRow(**data))
    if bulk_insert:
        await PlayerRow.insert(*bulk_insert)


async def migrate_node_config_v_1_0_0(nodes: list[asyncpg.Record]) -> None:
    """Processes node config for migration to version 1.0.0."""
    for node in nodes:
        data = {
            NodeRow.name: node["name"],
            NodeRow.ssl: node["ssl"],
            NodeRow.resume_timeout: node["resume_timeout"],
            NodeRow.reconnect_attempts: node["reconnect_attempts"],
            NodeRow.search_only: node["search_only"],
            NodeRow.managed: node["managed"],
            NodeRow.disabled_sources: node["disabled_sources"],
            NodeRow.extras: json.loads(node["extras"]),
            NodeRow.yaml: json.loads(node["yaml"]),
        }
        node_obj = await NodeRow.objects().get_or_create(NodeRow.id == node["id"], defaults=data)
        if not node_obj._was_created:
            await NodeRow.update(data).where(NodeRow.id == node["id"])


async def migrate_lib_config_v_1_0_0(configs: list[asyncpg.Record]) -> None:
    """Processes lib config for migration to version 1.0.0."""
    for config in configs:
        data = {
            LibConfigRow.config_folder: config["config_folder"],
            LibConfigRow.java_path: config["java_path"],
            LibConfigRow.enable_managed_node: config["enable_managed_node"],
            LibConfigRow.auto_update_managed_nodes: config["auto_update_managed_nodes"],
            LibConfigRow.localtrack_folder: config["localtrack_folder"],
            LibConfigRow.download_id: config["download_id"],
            LibConfigRow.update_bot_activity: config["update_bot_activity"],
            LibConfigRow.use_bundled_pylav_external: config["use_bundled_pylav_external"],
            LibConfigRow.use_bundled_lava_link_external: False,
            LibConfigRow.extras: json.loads(config["extras"]),
            LibConfigRow.next_execution_update_bundled_playlists: config["next_execution_update_bundled_playlists"],
            LibConfigRow.next_execution_update_bundled_external_playlists: config[
                "next_execution_update_bundled_external_playlists"
            ],
            LibConfigRow.next_execution_update_external_playlists: config["next_execution_update_external_playlists"],
        }
        config_obj = await LibConfigRow.objects().get_or_create(
            (LibConfigRow.id == config["id"]) & (LibConfigRow.bot == config["bot"]), defaults=data
        )
        if not config_obj._was_created:
            await LibConfigRow.update(data).where(
                (LibConfigRow.id == config["id"]) & (LibConfigRow.bot == config["bot"])
            )


async def low_level_v_1_0_0_migration(
    con: Connection, migration_data: dict[str, dict[str, list[asyncpg.Record]] | None], migrator: ConfigController
) -> None:
    """Runs the low level migration for version 1.0.0."""
    version = "1.0.0"
    migration_data[version] = {}
    await low_level_v_1_0_0_playlists(con, migration_data, version)
    await low_level_v_1_0_0_queries(con, migration_data, version)
    await low_level_v_1_0_0_players(con, migration_data, version)
    await low_level_v_1_0_0_lib(con, migration_data, version)
    await low_level_v_1_0_0_nodes(con, migration_data, version)
    if migration_data[version]:
        await migrator.reset_database()


async def low_level_v_1_0_0_nodes(
    con: Connection, migration_data: dict[str, dict[str, list[asyncpg.Record]] | None], version: str
) -> None:
    """Runs the low level migration of nodes  for version 1.0.0."""
    node_config_data_1000 = await run_node_config_v_1_0_0(con)
    if node_config_data_1000:
        migration_data[version]["node"] = node_config_data_1000


async def low_level_v_1_0_0_lib(
    con: Connection, migration_data: dict[str, dict[str, list[asyncpg.Record]] | None], version: str
) -> None:
    """Runs the low level migration of lib config for version 1.0.0."""
    lib_config_data_1000 = await run_lib_config_v_1_0_0(con)
    if lib_config_data_1000:
        migration_data[version]["lib"] = lib_config_data_1000


async def low_level_v_1_0_0_players(
    con: Connection, migration_data: dict[str, dict[str, list[asyncpg.Record]] | None], version: str
) -> None:
    """Runs the low level migration of players for version 1.0.0."""
    player_data_1000 = await run_player_config_v_1_0_0(con)
    if player_data_1000:
        migration_data[version]["player"] = player_data_1000


async def low_level_v_1_0_0_queries(
    con: Connection, migration_data: dict[str, dict[str, list[asyncpg.Record]] | None], version: str
) -> None:
    """Runs the low level migration of queries for version 1.0.0."""
    query_data_1000 = await run_query_migration_v_1_0_0(con)
    if query_data_1000:
        migration_data[version]["query"] = query_data_1000


async def low_level_v_1_0_0_playlists(
    con: Connection, migration_data: dict[str, dict[str, list[asyncpg.Record]] | None], version: str
) -> None:
    """Runs the low level migration of playlists for version 1.0.0."""
    playlist_data_1000 = await run_playlist_migration_v_1_0_0(con)
    if playlist_data_1000:
        migration_data[version]["playlist"] = playlist_data_1000
